#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
jinja-based :class:`~nunavut.generators.AbstractGenerator` implementation.
"""

import abc
import datetime
import io
import logging
import re
import shutil
from pathlib import Path
from typing import Any, Callable, Dict, Generator, Iterable, List, Mapping, Optional, TextIO, Tuple, Type, Union

import pydsdl

import nunavut.lang

from .._generators import AbstractGenerator
from .._postprocessors import FilePostProcessor, LinePostProcessor, PostProcessor
from .._utilities import TEMPLATE_SUFFIX, ResourceSearchPolicy, ResourceType, YesNoDefault
from .environment import CodeGenEnvironment, CodeGenEnvironmentBuilder
from .loaders import DEFAULT_TEMPLATE_PATH, DSDLSupportTemplateLoader, DSDLTemplateLoader

logger = logging.getLogger(__name__)

# +---------------------------------------------------------------------------+
# | JINJA : CodeGenerator
# +---------------------------------------------------------------------------+


class CodeGenerator(AbstractGenerator):
    """
    Abstract base class for all Generators that build source code using Jinja templates.

    :param int resource_types:             A bitfield of :class:`nunavut._utilities.ResourceType` for filtering the
                                           types of resources this generator will emit.
    :param nunavut.Namespace namespace:    The top-level namespace to generates code
                                           at and from.
    :param YesNoDefault generate_namespace_types:  Set to YES to emit files for namespaces.
                                           NO will suppress namespace file generation and DEFAULT will
                                           use the language's preference.
    :param templates_dir:                  Directories containing jinja templates. These will be available along
                                           with any built-in templates provided by the target language. The templates
                                           at these paths will take precedence masking any built-in templates
                                           where the names are the same. See :class:`jinja2.ChoiceLoader` for rules
                                           on the lookup hierarchy.
    :type templates_dir: Optional[Union[Path,List[Path]]]
    :param Optional[Type[DSDLTemplateLoader]] template_loader:
        If None uses an internal default implementation of :class:`DSDLTemplateLoader` otherwise instantiates the given
        class with the arguments specified by the :class:`DSDLTemplateLoader` constructor.
    :param bool followlinks:               If True then symbolic links will be followed when
                                           searching for templates.
    :param bool trim_blocks:               If this is set to True the first newline after a
                                           block is removed (block, not variable tag!).
    :param bool lstrip_blocks:             If this is set to True leading spaces and tabs
                                           are stripped from the start of a line to a block.
                                           Defaults to False.
    :param Dict[str, Callable] additional_filters: Optional jinja filters to add to the
                                           global environment using the key as the filter name
                                           and the callable as the filter.
    :param Dict[str, Callable] additional_tests: Optional jinja tests to add to the
                                           global environment using the key as the test name
                                           and the callable as the test.
    :param Dict[str, Any] additional_globals: Optional objects to add to the template
                                            environment globals collection.
    :param post_processors: A list of :class:`nunavut.postprocessors.PostProcessor`
    :type post_processors: Optional[List[nunavut.postprocessors.PostProcessor]]
    :param builtin_template_path: If provided overrides the folder name under which built-in templates are loaded from
                                            within a target language's package (i.e. ignored if no target language is
                                            specified). For example, if the target language is ``c`` and this parameter
                                            was set to ``foo`` then built-in templates would be loaded from
                                            ``nunavut.lang.c.foo``.
    :type builtin_template_path: str
    :param search_policy: The policy to use when searching for templates.
    :type search_policy: ResourceSearchPolicy
    :param embed_auditing_info: If True then the generator will embed auditing information in the generated code.
    :type embed_auditing_info: bool
    :raises RuntimeError: If any additional filter or test attempts to replace a built-in
                          or otherwise already defined filter or test.
    """

    @staticmethod
    def __augment_post_processors_with_ln_limit_empty_lines(
        post_processors: Optional[List[PostProcessor]], limit_empty_lines: int
    ) -> List[PostProcessor]:
        """
        Subroutine of _handle_post_processors method.
        """
        from nunavut._postprocessors import LimitEmptyLines  # pylint: disable=import-outside-toplevel

        if post_processors is None:
            post_processors = [LimitEmptyLines(limit_empty_lines)]
        else:
            found_pp = False
            for pp in post_processors:
                if isinstance(pp, LimitEmptyLines):
                    found_pp = True
                    break
            if not found_pp:
                post_processors.append(LimitEmptyLines(limit_empty_lines))
        return post_processors

    @staticmethod
    def __augment_post_processors_with_ln_trim_trailing_whitespace(
        post_processors: Optional[List[PostProcessor]],
    ) -> List[PostProcessor]:
        """
        Subroutine of _handle_post_processors method.
        """
        from nunavut._postprocessors import TrimTrailingWhitespace  # pylint: disable=import-outside-toplevel

        if post_processors is None:
            post_processors = [TrimTrailingWhitespace()]
        else:
            found_pp = False
            for pp in post_processors:
                if isinstance(pp, TrimTrailingWhitespace):
                    found_pp = True
                    break
            if not found_pp:
                post_processors.append(TrimTrailingWhitespace())
        return post_processors

    @classmethod
    def _handle_post_processors(
        cls,
        target_language: "nunavut.lang.Language",
        post_processors: Optional[List["PostProcessor"]],
    ) -> Optional[List["PostProcessor"]]:
        """
        Used by constructor to process an optional list of post-processors and to augment or create this list
        if needed to support language options.
        """
        try:
            limit_empty_lines = target_language.get_config_value("limit_empty_lines")
            post_processors = cls.__augment_post_processors_with_ln_limit_empty_lines(
                post_processors, int(limit_empty_lines)
            )
        except KeyError:
            pass

        if target_language.get_config_value_as_bool("trim_trailing_whitespace"):
            post_processors = cls.__augment_post_processors_with_ln_trim_trailing_whitespace(post_processors)

        return post_processors

    def __init__(
        self,
        namespace: nunavut.Namespace,
        resource_types: int = ResourceType.ANY.value,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
        templates_dir: Optional[Union[Path, List[Path]]] = None,
        template_loader: Optional[Type[DSDLTemplateLoader]] = None,
        followlinks: bool = False,
        trim_blocks: bool = False,
        lstrip_blocks: bool = False,
        additional_filters: Optional[Dict[str, Callable]] = None,
        additional_tests: Optional[Dict[str, Callable]] = None,
        additional_globals: Optional[Dict[str, Any]] = None,
        post_processors: Optional[List["PostProcessor"]] = None,
        builtin_template_path: str = DEFAULT_TEMPLATE_PATH,
        search_policy: ResourceSearchPolicy = ResourceSearchPolicy.FIND_ALL,
        embed_auditing_info: bool = False,
        **kwargs: Any,
    ):
        super().__init__(namespace, resource_types, generate_namespace_types, **kwargs)

        if templates_dir is not None and not isinstance(templates_dir, list):
            templates_dir = [templates_dir]

        language_context = self._namespace.get_language_context()
        target_language = language_context.get_target_language()

        if template_loader is None:
            template_loader = DSDLTemplateLoader

        self._dsdl_template_loader = template_loader(
            namespace=namespace,
            resource_types=resource_types,
            templates_dirs=templates_dir,
            followlinks=followlinks,
            builtin_template_path=builtin_template_path,
            search_policy=search_policy,
            **kwargs,
        )

        self._post_processors = self._handle_post_processors(target_language, post_processors)

        env_builder = (
            CodeGenEnvironmentBuilder(self._dsdl_template_loader)
            .set_trim_blocks(trim_blocks)
            .set_lstrip_blocks(lstrip_blocks)
        )
        if additional_filters is not None:
            env_builder.add_filters(**additional_filters)
        if additional_tests is not None:
            env_builder.add_tests(**additional_tests)
        if additional_globals is not None:
            env_builder.add_globals(**additional_globals)
        env_builder.set_embed_auditing_info(embed_auditing_info)

        self._env = self._create_environment(env_builder, language_context)

    def _create_environment(
        self, env_builder: CodeGenEnvironmentBuilder, language_context: nunavut.lang.LanguageContext
    ) -> CodeGenEnvironment:
        """
        Create the code generation environment for this generator.

        This is a template method that subclasses can override to customize the environment
        with generator-specific filters, tests, and globals. The base implementation creates
        an environment with language-specific filters only.

        :param env_builder: The environment builder with basic configuration already set.
        :param language_context: The language context for this generator.
        :return: A configured CodeGenEnvironment instance.
        """
        return env_builder.create(language_context)

    @property
    def dsdl_loader(self) -> DSDLTemplateLoader:
        """
        The template loader used by this generator.
        """
        return self._dsdl_template_loader

    @property
    def language_context(self) -> nunavut.lang.LanguageContext:
        """
        The language context used by this generator.
        """
        return self._namespace.get_language_context()

    @property
    def environment(self) -> CodeGenEnvironment:
        """
        The generator environment.
        """
        return self._env

    # +-----------------------------------------------------------------------+
    # | PROTECTED
    # +-----------------------------------------------------------------------+
    def _handle_overwrite(self, output_path: Path, allow_overwrite: bool) -> None:
        if output_path.exists():
            if allow_overwrite:
                output_path.chmod(output_path.stat().st_mode | 0o220)
            else:
                raise PermissionError("{output_path} exists and allow_overwrite is False.")

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+

    def get_templates(self) -> Iterable[Path]:
        """
        Enumerate all templates found in the templates path.
        :data:`~TEMPLATE_SUFFIX` as the suffix for the filename.

        :return: A list of paths to all templates found by this Generator object.
        """
        return self._dsdl_template_loader.get_templates()

    @abc.abstractmethod
    def generate_all(
        self,
        is_dryrun: bool = False,
        allow_overwrite: bool = True,
    ) -> Iterable[Path]:
        raise NotImplementedError()

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    @staticmethod
    def _filter_and_write_line(
        line_and_lineend: Tuple[str, str],
        output_file: TextIO,
        line_pps: List["LinePostProcessor"],
    ) -> None:
        for line_pp in line_pps:
            line_and_lineend = line_pp(line_and_lineend)
            if line_and_lineend is None:
                raise ValueError(
                    "line post processor must return a 2-tuple. To elide a line return a tuple of empty"
                    "strings. None is not a valid value."
                )

        output_file.write(line_and_lineend[0])
        output_file.write(line_and_lineend[1])

    @classmethod
    def _generate_with_line_buffer(
        cls,
        output_file: TextIO,
        template_gen: Generator[str, None, None],
        line_pps: List["LinePostProcessor"],
    ) -> None:
        newline_pattern = re.compile(r"\n|\r\n", flags=re.MULTILINE)
        line_buffer = io.StringIO()
        for part in template_gen:
            search_pos = 0  # type: int
            match_obj = newline_pattern.search(part, search_pos)
            while True:
                if search_pos < 0 or search_pos >= len(part):
                    break
                if match_obj is None:
                    line_buffer.write(part[search_pos:])
                    break

                # We have a newline
                line_buffer.write(part[search_pos : match_obj.start()])
                newline_chars = part[match_obj.start() : match_obj.end()]
                line = line_buffer.getvalue()  # type: str
                line_buffer = io.StringIO()
                cls._filter_and_write_line((line, newline_chars), output_file, line_pps)
                search_pos = match_obj.end()
                match_obj = newline_pattern.search(part, search_pos)
        remainder = line_buffer.getvalue()
        if len(remainder) > 0:
            cls._filter_and_write_line((remainder, ""), output_file, line_pps)

    def _generate_code(
        self,
        output_path: Path,
        template_gen: Generator[str, None, None],
        allow_overwrite: bool,
    ) -> None:
        """
        Logic that should run from _generate_type iff is_dryrun is False.
        """

        try:
            self._env.now_utc = datetime.datetime.now(datetime.UTC)  # type: ignore
        except AttributeError:
            self._env.now_utc = datetime.datetime.utcnow()

        from ..lang._common import UniqueNameGenerator  # pylint: disable=import-outside-toplevel

        # reset the name generator state for this type
        UniqueNameGenerator.reset()

        # Predetermine the post processor types.
        line_pps = []  # type: List['LinePostProcessor']
        file_pps = []  # type: List['FilePostProcessor']
        if self._post_processors is not None:
            for pp in self._post_processors:
                if isinstance(pp, LinePostProcessor):
                    line_pps.append(pp)
                elif isinstance(pp, FilePostProcessor):
                    file_pps.append(pp)
                else:
                    raise ValueError(f"PostProcessor type {type(pp)} is unknown.")
        logger.debug("Using post-processors: %r %r", line_pps, file_pps)

        self._handle_overwrite(output_path, allow_overwrite)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(output_path), "w", encoding="utf-8") as output_file:
            if len(line_pps) > 0:
                # The logic gets much more complex when doing line post-processing.
                self._generate_with_line_buffer(output_file, template_gen, line_pps)
            else:
                for part in template_gen:
                    output_file.write(part)
        for file_pp in file_pps:
            output_path = file_pp(output_path)


# +---------------------------------------------------------------------------+
# | JINJA : DSDLCodeGenerator
# +---------------------------------------------------------------------------+


class DSDLCodeGenerator(CodeGenerator):
    """
    :class:`~CodeGenerator` implementation that generates code for a given set
    of DSDL types.
    """

    # +-----------------------------------------------------------------------+
    # | JINJA : filters
    # +-----------------------------------------------------------------------+

    @staticmethod
    def filter_yamlfy(value: Any) -> str:
        """
        Filter to, optionally, emit a dump of the dsdl input as a yaml document.
        Available as ``yamlfy`` in all template environments.

        Example::

            /*
            {{ T | yamlfy }}
            */

        Result Example (truncated for brevity)::

            /*
            !!python/object:pydsdl.StructureType
            _attributes:
            - !!python/object:pydsdl.Field
            _serializable: !!python/object:pydsdl.UnsignedIntegerType
                _bit_length: 16
                _cast_mode: &id001 !!python/object/apply:pydsdl.CastMode
                - 0
            _name: value
            */

        :param value: The input value to parse as yaml.

        :return: If a yaml parser is available, a pretty dump of the given value as yaml.
                  If a yaml parser is not available then an empty string is returned.
        """
        from yaml import Dumper as YamlDumper  # pylint: disable=import-outside-toplevel
        from yaml import dump as yaml_dump  # pylint: disable=import-outside-toplevel

        return str(yaml_dump(value, Dumper=YamlDumper))

    def filter_type_to_template(self, value: Any) -> str:
        """
        Template for type resolution as a filter. Available as ``type_to_template``
        in all template environments.

        Example::

            {%- for attribute in T.attributes %}
                {%* include attribute.data_type | type_to_template %}
                {%- if not loop.last %},{% endif %}
            {%- endfor %}

        :param value: The input value to change into a template include path.

        :return: A path to a template named for the type with :data:`TEMPLATE_SUFFIX`
        """
        result = self.dsdl_loader.type_to_template(type(value))
        if result is None:
            raise RuntimeError(f"No template found for type {value}")
        return result.name

    def filter_type_to_include_path(self, value: Any, resolve: bool = False) -> str:
        """
        Emits an include path to the output target for a given type.

        Example::

            # include "{{ T.my_type | type_to_include_path }}"

        Result Example:

            # include "foo/bar/my_type.h"

        :param Any value: The type to emit an include for.
        :param bool resolve: If True the path returned will be absolute else the path will
                             be relative to the folder of the root namespace.
        :return: A string path to output file for the type.
        """

        include_path = self.namespace.find_output_path_for_type(value)
        if resolve:
            return include_path.resolve().as_posix()
        else:
            return include_path.relative_to(self.namespace.output_folder.parent).as_posix()

    @staticmethod
    def filter_typename(value: Any) -> str:
        """
        Filters a given token as its type name. Available as ``typename``
        in all template environments.

        This example supposes that ``T.some_value == "some string"``

        Example::

            {{ T.some_value | typename }}

        Result Example::

            str

        :param value: The input value to filter into a type name.

        :return: The ``__name__`` of the python type.
        """
        return type(value).__name__

    @staticmethod
    def filter_alignment_prefix(offset: pydsdl.BitLengthSet) -> str:
        """
        Provides a string prefix based on a given :class:`pydsdl.BitLengthSet`.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

        .. code-block:: python

            # Given
            B = pydsdl.BitLengthSet(32)

            # and
            template = '{{ B | alignment_prefix }}'

            # outputs
            rendered = 'aligned'

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_alignment_prefix, template, rendered, 'py', B=B)


        .. code-block:: python

            # Given
            B = pydsdl.BitLengthSet(32)
            B += 1

            # and
            template = '{{ B | alignment_prefix }}'

            # outputs
            rendered = 'unaligned'

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_alignment_prefix, template, rendered, 'py', B=B)


        :param pydsdl.BitLengthSet offset: A bit length set to test for alignment.
        :return: 'aligned' or 'unaligned' based on the state of the ``offset`` argument.
        """
        if isinstance(offset, pydsdl.BitLengthSet):
            return "aligned" if offset.is_aligned_at_byte() else "unaligned"
        else:  # pragma: no cover
            raise TypeError(f"Expected BitLengthSet, got {type(offset).__name__}")

    @staticmethod
    def filter_bit_length_set(values: Optional[Union[Iterable[int], int]]) -> pydsdl.BitLengthSet:
        """
        Convert an integer or a list of integers into a :class:`pydsdl.BitLengthSet`.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

            assert type(DSDLCodeGenerator.filter_bit_length_set(23)) == pydsdl.BitLengthSet

        """
        return pydsdl.BitLengthSet(values)

    @staticmethod
    def filter_remove_blank_lines(text: str) -> str:
        """
        Remove blank lines from the supplied string.
        Lines that contain only whitespace characters are also considered blank.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

        .. code-block:: python

            # Given
            text = '''123

            456
            \t
            \v\f
            789'''

            # and
            template = '{{ text | remove_blank_lines }}'

            # then the black lines will be removed leaving...
            rendered = '''123
            456
            789'''

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_remove_blank_lines, template, rendered, 'c', text=text)

        """
        return re.sub(r"\n([ \t\f\v]*\n)+", r"\n", text)

    @staticmethod
    def filter_bits2bytes_ceil(n_bits: int) -> int:
        """
        Implements ``int(ceil(x/8)) | x >= 0``.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(50) == 7
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(8) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(7) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(1) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(0) == 0

        """
        if n_bits < 0:
            raise ValueError("The number of bits cannot be negative")
        return (int(n_bits) + 7) // 8

    @staticmethod
    def filter_text_table(data: Dict, start_each_line: str, column_sep: str = " : ", line_end: str = "\n") -> str:
        """
        Create a text table from a dictionary of data.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

        .. code-block:: python

            # Given
            table = {
                "banana": "yellow",
                "apple": "red",
                "grape": "purple"
            }

            # and
            template = '''
            {{ table | text_table("//  ", " | ", "\\n") }}'''

            # then
            rendered = '''
            //  banana | yellow
            //  apple  | red
            //  grape  | purple'''

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_text_table, template, rendered, 'c', table=table)

        """
        # Find the longest key to set the width of the first column
        key_width = max(len(key) for key in data.keys())

        output = []
        for key, value in data.items():
            output.append(f"{start_each_line}{key:<{key_width}}{column_sep}{value}".rstrip())
        return line_end.join(output)

    # +-----------------------------------------------------------------------+
    # | JINJA : tests
    # +-----------------------------------------------------------------------+

    @staticmethod
    def is_None(value: Any) -> bool:  # pylint: disable=invalid-name
        """
        Tests if a value is ``None``

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            assert DSDLCodeGenerator.is_None(None) is True
            assert DSDLCodeGenerator.is_None(1) is False

        """
        return value is None

    @staticmethod
    def is_saturated(t: pydsdl.PrimitiveType) -> bool:
        """
        Tests if a type is a saturated type or not.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl
            import pytest

            saturated_mock = MagicMock(spec=pydsdl.PrimitiveType)
            saturated_mock.cast_mode = pydsdl.PrimitiveType.CastMode.SATURATED
            assert DSDLCodeGenerator.is_saturated(saturated_mock) is True

            truncated_mock = MagicMock(spec=pydsdl.PrimitiveType)
            truncated_mock.cast_mode = pydsdl.PrimitiveType.CastMode.TRUNCATED
            assert DSDLCodeGenerator.is_saturated(truncated_mock) is False

            with pytest.raises(TypeError):
                 DSDLCodeGenerator.is_saturated(MagicMock(spec=pydsdl.SerializableType))

        """
        if isinstance(t, pydsdl.PrimitiveType):
            return {
                pydsdl.PrimitiveType.CastMode.SATURATED: True,
                pydsdl.PrimitiveType.CastMode.TRUNCATED: False,
            }[t.cast_mode]
        else:
            raise TypeError(f"Cast mode is not defined for {type(t).__name__}")

    @staticmethod
    def is_service_request(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is request type of a service type.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            service_request_mock = MagicMock(spec=pydsdl.SerializableType)
            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Request'
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is True

            service_request_mock.has_parent_service = False
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is False

            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Response'
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is False

        """
        return instance.has_parent_service and instance.full_name.split(".")[-1] == "Request"  # type: ignore

    @staticmethod
    def is_service_response(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is response type of a service type.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            service_request_mock = MagicMock(spec=pydsdl.SerializableType)
            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Response'
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is True

            service_request_mock.has_parent_service = False
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is False

            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Request'
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is False

        """
        return instance.has_parent_service and instance.full_name.split(".")[-1] == "Response"  # type: ignore

    @staticmethod
    def is_deprecated(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is marked as deprecated

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            composite_type_mock = MagicMock(spec=pydsdl.CompositeType)
            composite_type_mock.deprecated = True
            assert DSDLCodeGenerator.is_deprecated(composite_type_mock) is True

            array_type_mock = MagicMock(spec=pydsdl.ArrayType)
            array_type_mock.element_type = composite_type_mock
            assert DSDLCodeGenerator.is_deprecated(array_type_mock) is True

            other_type_mock = MagicMock(spec=pydsdl.SerializableType)
            assert DSDLCodeGenerator.is_deprecated(other_type_mock) is False

        """
        if isinstance(instance, pydsdl.CompositeType):
            return instance.deprecated  # type: ignore
        elif isinstance(instance, pydsdl.ArrayType) and isinstance(instance.element_type, pydsdl.CompositeType):
            return instance.element_type.deprecated  # type: ignore
        else:
            return False

    # +-----------------------------------------------------------------------+

    def __init__(self, namespace: nunavut.Namespace, resource_types: int = ResourceType.ANY.value, **kwargs: Any):
        super().__init__(namespace, resource_types=resource_types, **kwargs)

    def _create_environment(
        self, env_builder: CodeGenEnvironmentBuilder, language_context: nunavut.lang.LanguageContext
    ) -> CodeGenEnvironment:
        """
        Create the environment with DSDL-specific filters and tests.

        This override adds DSDL-specific functionality like type_to_template, type_to_include_path,
        and DSDL tests (is_service_request, is_service_response, etc.) that should only be
        available in DSDL type templates, not in support templates.

        :param env_builder: The environment builder with basic configuration already set.
        :param language_context: The language context for this generator.
        :return: A configured CodeGenEnvironment with DSDL-specific additions.
        """
        env = super()._create_environment(env_builder, language_context)

        # Add DSDL-specific tests (is_service_request, is_service_response, etc.)
        for test_name, test in self._create_all_dsdl_tests().items():
            env.add_test(test_name, test)

        # Add DSDL-specific filters (type_to_template, type_to_include_path, etc.)
        env.add_conventional_methods_to_environment(self)

        return env

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+
    def get_templates(self) -> Iterable[Path]:
        if (self.resource_types & ResourceType.ONLY.value) == ResourceType.ONLY.value:
            # This generator doesn't generate resources and the "only resources" flag is set.
            return []
        return super().get_templates()

    def generate_all(
        self,
        is_dryrun: bool = False,
        allow_overwrite: bool = True,
    ) -> Iterable[Path]:
        generated = []  # type: List[Path]
        if (self.resource_types & ResourceType.ONLY.value) == ResourceType.ONLY.value:
            # This generator doesn't generate resources and the "only resources" flag is set.
            # We'll set the dryrun flag to True to avoid generating anything but we'll still return the list of
            # types that would have been generated.
            is_dryrun = True
        provider = self.namespace.get_all_types if self.generate_namespace_types else self.namespace.get_all_datatypes
        for parsed_type, output_path in provider():
            logger.info("Generating: %s", parsed_type)
            generated.append(self._generate_type(parsed_type, output_path, is_dryrun, allow_overwrite))

        generated.extend(self._generate_index_files(is_dryrun, allow_overwrite))
        return generated

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    @classmethod
    def _create_instance_tests_for_type(cls, root: pydsdl.Any) -> Dict[str, Callable]:
        tests = dict()

        def _field_is_instance(field_or_datatype: pydsdl.Any) -> bool:
            if isinstance(field_or_datatype, pydsdl.Attribute):
                return isinstance(field_or_datatype.data_type, root)
            else:
                return isinstance(field_or_datatype, root)

        tests[root.__name__] = _field_is_instance
        root_name_lower = root.__name__.lower()
        if len(root_name_lower) > 4 and root_name_lower.endswith("type"):
            tests[root_name_lower[:-4]] = _field_is_instance
        elif len(root_name_lower) > 5 and root_name_lower.endswith("field"):
            tests[root_name_lower[:-5]] = _field_is_instance
        else:
            tests[root_name_lower] = _field_is_instance

        for derived in root.__subclasses__():
            tests.update(cls._create_instance_tests_for_type(derived))
        return tests

    @classmethod
    def _create_all_dsdl_tests(cls) -> Mapping[str, Callable]:
        """
        Create a collection of jinja tests for all base dsdl types.

        .. invisible-code-block: python

            import pydsdl
            from unittest.mock import MagicMock
            from nunavut.jinja import DSDLCodeGenerator

            test_set = DSDLCodeGenerator._create_all_dsdl_tests()

            def _do_pydsdl_instance_test_test(pydsdl_obj, test_name):

                if not test_set[test_name](pydsdl_obj):
                    raise AssertionError(test_name)

            def _do_pydsdl_instance_test_tests(pydsdl_type):
                mock_instance = MagicMock(spec=pydsdl_type)
                _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__)
                if pydsdl_type.__name__.endswith('Type'):
                    _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__[:-4].lower())
                if pydsdl_type.__name__.endswith('Field'):
                    _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__[:-5].lower())
                mock_attribute = MagicMock(spec=pydsdl.Attribute)
                mock_attribute.data_type = mock_instance
                _do_pydsdl_instance_test_test(mock_attribute, pydsdl_type.__name__)

            _do_pydsdl_instance_test_tests(pydsdl.SerializableType)
            _do_pydsdl_instance_test_tests(pydsdl.PrimitiveType)
            _do_pydsdl_instance_test_tests(pydsdl.IntegerType)
            _do_pydsdl_instance_test_tests(pydsdl.ServiceType)

        """
        all_tests = dict()
        all_tests.update(cls._create_instance_tests_for_type(pydsdl.SerializableType))
        all_tests.update(cls._create_instance_tests_for_type(pydsdl.Attribute))
        return all_tests

    def _generate_type(
        self, input_type: pydsdl.CompositeType, output_path: Path, is_dryrun: bool, allow_overwrite: bool
    ) -> Path:
        template_name = self.filter_type_to_template(input_type)
        template = self._env.get_template(template_name)
        template_gen = template.generate(T=input_type)
        if not is_dryrun:
            self._generate_code(output_path, template_gen, allow_overwrite)
        return output_path

    def _generate_index_files(
        self,
        is_dryrun: bool,
        allow_overwrite: bool,
    ) -> List[Path]:
        """
        Renders index files which are each given access to all namespaces as `N`.
        """
        output_paths = []
        index = self.namespace.get_index_namespace()
        index_file_path = index.output_folder
        target_extension = self.language_context.get_target_language().get_config_value(
            nunavut.lang.Language.WKCV_DEFINITION_FILE_EXTENSION
        )
        for index_file in self.index_files:
            template_name = self.dsdl_loader.index_file_to_template(index_file)
            if template_name is None:
                raise RuntimeError(f"No template found for index file {index_file}")
            template = self._env.get_template(template_name.name)
            template_gen = template.generate(N=index)
            index_file_output = index_file_path / index_file
            if len(index_file.suffix) == 0:
                index_file_output = index_file_output.with_suffix(target_extension)
            output_paths.append(index_file_output)
            if not is_dryrun:
                self._generate_code(index_file_output, template_gen, allow_overwrite)
        return output_paths


# +---------------------------------------------------------------------------+
# | JINJA : SupportGenerator
# +---------------------------------------------------------------------------+


class SupportGenerator(CodeGenerator):
    """
    Generates output files by copying them from within the Nunavut package itself
    for non templates but uses jinja to generate headers from templates with the
    language environment provided but no ``T`` (DSDL type) global set.
    This generator always copies files from those returned by the ``file_iterator``
    to locations under :func:`nunavut.Namespace.get_support_output_folder()`
    """

    def __init__(
        self,
        namespace: nunavut.Namespace,
        resource_types: int,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
        templates_dir: Optional[Union[Path, List[Path]]] = None,
        **kwargs: Any,
    ):
        super().__init__(
            namespace,
            resource_types,
            generate_namespace_types=generate_namespace_types,
            templates_dir=templates_dir,
            builtin_template_path="support",
            template_loader=DSDLSupportTemplateLoader,
            **kwargs,
        )

        target_language = self.language_context.get_target_language()

        #  Create the sub-folder to copy-to based on the support namespace.
        self._sub_folders = Path("")

        for namespace_part in target_language.support_namespace:
            self._sub_folders = self._sub_folders / Path(namespace_part)

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+
    def get_templates(self) -> Iterable[Path]:
        if self.resource_types == 0:
            # This generator only generates resources and the "no resources" flag is set.
            return []
        return super().get_templates()

    def generate_all(
        self,
        is_dryrun: bool = False,
        allow_overwrite: bool = True,
    ) -> Iterable[Path]:
        generated = []  # type: List[Path]
        if self.resource_types == 0:
            # This generator only generates resources and the "no resources" flag is set.
            return generated

        target_language = self.language_context.get_target_language()
        target_path = Path(self.namespace.get_index_namespace().base_output_path) / self._sub_folders

        line_pps: List[LinePostProcessor] = []
        file_pps: List[FilePostProcessor] = []
        if self._post_processors is not None:
            for pp in self._post_processors:
                if isinstance(pp, LinePostProcessor):
                    line_pps.append(pp)
                elif isinstance(pp, FilePostProcessor):
                    file_pps.append(pp)
                else:
                    raise ValueError(f"PostProcessor type {type(pp)} is unknown.")

        for resource in self.get_templates():
            target = (target_path / resource.name).with_suffix(target_language.extension)
            logger.info("Generating support file: %s", target)
            if resource.suffix == TEMPLATE_SUFFIX:
                self._generate_header(resource, target, is_dryrun, allow_overwrite)
                generated.append(target)
            else:
                self._copy_header(resource, target, is_dryrun, allow_overwrite, line_pps, file_pps)
                generated.append(target)
        return generated

    # +-----------------------------------------------------------------------+
    # | Private
    # +-----------------------------------------------------------------------+

    def _generate_header(self, template_path: Path, output_path: Path, is_dryrun: bool, allow_overwrite: bool) -> Path:
        template = self._env.get_template(template_path.name)
        template_gen = template.generate()
        if not is_dryrun:
            self._generate_code(output_path, template_gen, allow_overwrite)
        return output_path

    def _copy_header(
        self,
        resource: Path,
        target: Path,
        is_dryrun: bool,
        allow_overwrite: bool,
        line_pps: List["LinePostProcessor"],
        file_pps: List["FilePostProcessor"],
    ) -> Path:
        if not is_dryrun:
            self._handle_overwrite(target, allow_overwrite)
            target.parent.mkdir(parents=True, exist_ok=True)
            if len(line_pps) == 0:
                shutil.copy(str(resource), str(target))
            else:
                self._copy_header_using_line_pps(resource, target, line_pps)
            for file_pp in file_pps:
                target = file_pp(target)
        return target

    def _copy_header_using_line_pps(
        self,
        resource: Path,
        target: Path,
        line_pps: List["LinePostProcessor"],
    ) -> None:
        with open(str(target), "w", encoding="utf-8") as target_file:
            with open(str(resource), "r", encoding="utf-8") as resource_file:
                for resource_line in resource_file:
                    if len(resource_line) > 1 and resource_line[-2] == "\r":
                        resource_line_tuple = (resource_line[0:-2], "\r\n")
                    else:
                        resource_line_tuple = (resource_line[0:-1], "\n")
                    for line_pp in line_pps:
                        resource_line_tuple = line_pp(resource_line_tuple)
                    target_file.write(resource_line_tuple[0])
                    target_file.write(resource_line_tuple[1])



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Contains the Jinja templates to generate javascript.
"""

__version__ = "1.0.0"
"""Version of the js template set."""



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Namespace object and associated utilities. Nunavut namespaces provide an internal representation of dsdl namespaces
that are also data objects for target languages, like python, that model namespaces as objects.

The one Namespace type can play three roles:

1. **Namespace Index**: The root of a tree of namespaces. This is a special namespace that is used to generate index
    files that reference all the generated files in the tree at once. It is always the parent of each root namespace.
    Index namespaces contain only root namespaces and have no data types. They make no assumptions about the structure
    of the namespaces they contain.
2. **Root Namespace**: A namespace that is a direct child of the index namespace. It contains the namespaces's datatypes
    and nested namespaces.
3. **Nested Namespace**: A namespace that is a child of another namespace. It contains the namespaces's datatypes and
    nested namespaces.

```
        ┌─────────────┐
        │  Namespace  │
        │   «index»   │ (the index cannot contain data types)
        │             │
        └──────┬──────┘
               │1
               │
               │
               │*
               ▼
        ┌─────────────┐            ┌──────────┐
        │  Namespace  │1          *│ DataType │
        │   «root»    ├───────────►│          │
        │             │            └──────────┘
        └──────┬──────┘
               │1
               │
               │
               │*
               ▼
        ┌─────────────┐            ┌──────────┐
        │  Namespace  │1          *│ DataType │
        │  «nested»   ├───────────►│          │
        │             │            └──────────┘
        └─────────────┘
```
"""

import collections
import itertools
import logging
import multiprocessing
import multiprocessing.pool
import sys
from functools import singledispatchmethod
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Deque,
    Generator,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Protocol,
    Tuple,
    TypeVar,
    cast,
)

import pydsdl

from .lang import Language, LanguageContext
from .lang._common import IncludeGenerator

# +--------------------------------------------------------------------------------------------------------------------+


class AsyncResultProtocol(Protocol):
    """
    Defines the protocol for a duck-type compatible with multiprocessing.pool.AsyncResult.
    """

    def get(self, timeout: Any | None = None) -> Any:
        """
        See multiprocessing.pool.AsyncResult.get
        """


class NotAsyncResult:
    """
    Duck-type compatible with multiprocessing.pool.AsyncResult that is not actually asynchronous. All work is performed
    synchronously when the get method is called.
    """

    def __init__(self, read_method: Callable[..., Any], args: Tuple[Any, ...]) -> None:
        self.read_method = read_method
        self.args = args
        self._logger = logging.getLogger(NotAsyncResult.__name__)

    def get(self, timeout: Any | None = None) -> Any:
        """
        Perform the work synchronously.
        """
        if timeout is not None and timeout > 0:
            self._logger.debug(
                "Timeout value for read_method '%s' ignored when not doing multiple jobs.", self.read_method.__name__
            )
        return self.read_method(*self.args)


ApplyMethodT = TypeVar("ApplyMethodT", bound=Callable[..., AsyncResultProtocol])


def _read_files_strategy(
    index: "Namespace",
    apply_method: ApplyMethodT,
    dsdl_files: Path | str | Iterable[Path | str],
    job_timeout_seconds: float,
    omit_dependencies: bool,
    args: Iterable[Any],
) -> "Namespace":
    """
    Strategy for reading a set of dsdl files and building a namespace tree. This strategy is compatible with both
    synchronous and asynchronous invocation of the pydsdl.read_files method.
    """
    if isinstance(dsdl_files, (str, Path)):
        fileset = {Path(dsdl_files)}
    else:
        fileset = {Path(file) for file in dsdl_files}
    resolve_cache: dict[Path, Path] = {}

    def _resolve_file(file: Path) -> Path:
        # limit filesystem access by caching resolved files. This assumes the non-canonical path form
        # is consistent. If not this would not limit filesystem access but would be correct.
        if file not in resolve_cache:
            resolve_cache[file] = file.resolve()
        return resolve_cache[file]

    running_lookups: list[AsyncResultProtocol] = []
    already_read: set[Path] = set()
    while fileset:
        next_file = fileset.pop()
        running_lookups.append(apply_method(pydsdl.read_files, args=itertools.chain([next_file], args)))
        already_read.add(_resolve_file(next_file))
        if not fileset:
            for lookup in running_lookups:
                if job_timeout_seconds <= 0:
                    target_type, dependent_types = lookup.get()
                else:
                    target_type, dependent_types = lookup.get(timeout=job_timeout_seconds)
                Namespace.add_types(index, (target_type[0], dependent_types))
                if not omit_dependencies:
                    for dependent_type in dependent_types:
                        if _resolve_file(dependent_type.source_file_path) not in already_read:
                            fileset.add(dependent_type.source_file_path)
            running_lookups.clear()

    return index


# +--------------------------------------------------------------------------------------------------------------------+
class Generatable(type(Path())):  # type: ignore
    """
    A file that can be generated from a pydsdl type.

    .. invisible-code-block: python

        from nunavut._namespace import Generatable
        from pathlib import Path
        from unittest.mock import MagicMock
        import pydsdl

        dsdl_definition = MagicMock(spec=pydsdl.CompositeType)
        dependent_types = [MagicMock(spec=pydsdl.CompositeType)]

    .. code-block:: python

        # Generatables combine a Path to the generated file with the pydsdl type that can be reified into the file
        # and the types that are required to generate the file. This is useful for tracking dependencies and
        # generating files in the correct order. It also provides a representation of the generated file before it
        # is actually generated.

        generatable = Generatable(dsdl_definition, dependent_types, "test.h")

        # This is a Generatable object.
        assert isinstance(generatable, Generatable)
        assert generatable.definition == dsdl_definition
        assert generatable.input_types == dependent_types

        # But it is also a Path object.
        assert isinstance(generatable, Path)
        assert Path("test.h") == generatable

    :param pydsdl.Any definition: The pydsdl type that can be reified into a generated file.
    :param List[pydsdl.Any] input_types: The types that are required to generate the file.
    :param args: Arguments to pass to the Path constructor.
    :param kwargs: Keyword arguments to pass to the Path constructor.
    """

    @classmethod
    def _check_arguments(
        cls, definition: pydsdl.CompositeType, input_types: List[pydsdl.CompositeType]
    ) -> Tuple[pydsdl.CompositeType, List[pydsdl.CompositeType]]:
        """
        Check the arguments for the Generatable constructor.

        :param pydsdl.Any definition: The pydsdl type that can be reified into a generated file.
        :param List[pydsdl.Any] input_types: The types that are required to generate the file.
        :raises TypeError: If the arguments are not of the correct types.
        :return: The definition and input types.
        """
        if not isinstance(definition, pydsdl.CompositeType):
            raise TypeError("Generatable requires a 'definition' argument of type pydsdl.CompositeType.")
        if not isinstance(input_types, list):
            raise TypeError("Generatable requires an 'input_types' argument of type List[pydsdl.CompositeType].")
        return definition, input_types

    if sys.version_info < (3, 12):

        def __new__(cls, *args: Any, **kwargs: Any) -> "Generatable":
            """
            The override of the __new__ operator is required until python 3.12.
            After that, the __init__ operator can be used.
            """
            if cls is not Generatable:
                raise TypeError("Unknown type passed to Generatable constructor.")

            if len(args) < 3:
                raise TypeError("Generatable requires 'definition', 'input_types', and 'path' arguments.")

            definition, input_types = cls._check_arguments(*args[:2])
            new_pure_path = cast(Generatable, super().__new__(cls, *args[2:], **kwargs))
            new_pure_path._definition = definition
            new_pure_path._input_types = input_types
            return new_pure_path

    else:

        def __init__(
            self, definition: pydsdl.CompositeType, input_types: List[pydsdl.CompositeType], *args: Any, **kwargs: Any
        ):
            super().__init__(*args, **kwargs)
            self._definition, self._input_types = self._check_arguments(definition, input_types)

    @classmethod
    def wrap(
        cls, path: Path, definition: pydsdl.CompositeType, input_types: List[pydsdl.CompositeType]
    ) -> "Generatable":
        """
        Create a Generatable object from a Path, a pydsdl type, and a list of pydsdl types in a Python-version agnostic
        way. This is useful for deferred construction of Generatable objects since __init__ is not available in
        the python 3.11 and earlier versions.

        :param Path path: The path to the generated file.
        :param pydsdl.Any definition: The pydsdl type that can be reified into a generated file.
        :param List[pydsdl.Any] input_types: The types that are required to generate the file.
        :return: A Generatable object.
        """
        return Generatable(definition, input_types, path)

    def with_segments(self, *pathsegments: str | PathLike) -> Path:
        """
        Path override: Construct a new path object from any number of path-like objects.
        We discard the Generatable type here and continue on with a default Path object.
        """
        return Path(*pathsegments)

    @property
    def definition(self) -> pydsdl.CompositeType:
        """
        The pydsdl type that can be reified into a generated file.
        """
        return self._definition  # pylint: disable=no-member

    @property
    def input_types(self) -> List[pydsdl.CompositeType]:
        """
        The types that are required to generate the file.
        """
        return self._input_types.copy()  # type: ignore # pylint: disable=no-member,

    # --[DATA MODEL]-------------------------------------------------------------------------------------------------
    def __reduce__(self) -> Tuple[Callable, Tuple[Path, pydsdl.CompositeType, List[pydsdl.CompositeType]]]:
        super_reduction = super().__reduce__()
        reduced_path = Path(*super_reduction[1]) if isinstance(super_reduction, tuple) else Path(super_reduction)
        return (self.wrap, (reduced_path, self.definition, self.input_types))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Generatable):
            return bool(
                super().__eq__(other)
                and self._definition == other._definition  # pylint: disable=no-member
                and self._input_types == other._input_types  # pylint: disable=no-member
            )
        else:
            return super().__eq__(other)  # type: ignore

    def __hash__(self) -> int:
        return hash((super().__hash__(), self._definition))  # pylint: disable=no-member

    def __repr__(self) -> str:
        return (
            f"{super().__repr__()}, "  # pylint: disable=no-member
            f"definition={repr(self._definition)}, "  # pylint: disable=no-member
            f"input_types={repr(self._input_types)}"  # pylint: disable=no-member
        )

    def __copy__(self) -> "Generatable":
        return Generatable(self.definition, self.input_types, *self.parts)


# +--------------------------------------------------------------------------------------------------------------------+
class Namespace(pydsdl.Any):  # pylint: disable=too-many-public-methods
    """
    K-ary tree (where K is the largest set of data types in a single dsdl namespace) where
    the nodes represent dsdl namespaces and the children are the datatypes and other nested
    namespaces (with datatypes always being leaf nodes). This structure extends :code:`pydsdl.Any`
    and is a :code:`pydsdl.CompositeType` via duck typing.

    :param str full_namespace:  The full, dot-separated name of the namepace.
    :param Path namespace_dir: If full_namespace is "" then this is interpreted as the directory under which all
        namespaces and datatypes should be generated. Otherwise, it is the directory that contains the namespace's
        datatypes.
    :param LanguageContext language_context: The generated software language context the namespace is within.
    :param Namespace parent: The parent namespace of this namespace or None if this is an index namespace.
    """

    @classmethod
    def strop_namespace(cls, full_namespace: str, language_context: LanguageContext) -> Tuple[List[str], List[str]]:
        """
        Strop a namespace string for a given language context.

        :param str full_namespace: The dot-separated namespace string to strop.
        :param LanguageContext language_context: The language context to use when stroping the namespace.
        :return: A tuple containing the original namespace components and the stropped namespace components.

        .. invisible-code-block: python

            from nunavut.lang import LanguageContext, LanguageContextBuilder
            from nunavut._namespace import Namespace

            lctx = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .create()
            )

        .. code-block:: python

            full_namespace = "uavcan.node"
            namespace_components, namespace_components_stropped = Namespace.strop_namespace(full_namespace, lctx)

            assert namespace_components == ["uavcan", "node"]
            assert namespace_components_stropped == ["uavcan", "node"]

        """
        namespace_components = full_namespace.split(".")
        return (
            namespace_components,
            [language_context.filter_id_for_target(component, "path") for component in namespace_components],
        )

    @classmethod
    def add_types(
        cls,
        index: "Namespace",
        types: (
            Tuple[pydsdl.CompositeType, List[pydsdl.CompositeType]]
            | List[Tuple[pydsdl.CompositeType, List[pydsdl.CompositeType]]]
        ),
        extension: str | None = None,
    ) -> None:
        """
        Add a set of types to a namespace tree building new nodes as needed.

        :param Namespace tree: A namespace tree to add types to. This can be any namespace in the tree as
            :meth:`Namespace.get_index_namespace` will be used to find the meta-namespace.
        :param list types: A list of pydsdl types to add.
        :param str extension: Override the file extension to use for the generated files. If None, the extension will be
            determined by the target language.

        .. invisible-code-block: python


        """
        lctx = index.get_language_context()
        if not isinstance(types, list):
            types = [types]
        for dsdl_type, input_types in types:
            # For each type we form a path with the output_dir as the base; the intermediate
            # folders named for the type's namespaces; and a file name that includes the type's
            # short name, major version, minor version, and the extension argument as a suffix.
            # Python's pathlib adapts the provided folder and file names to the platform
            # this script is running on.
            # We also, lazily, generate Namespace nodes as we encounter new namespaces for the
            # first time.

            root_namespace_perhaps = index
            stropped, _ = cls.strop_namespace(dsdl_type.full_namespace, lctx)
            for namespace_component in stropped:
                if root_namespace_perhaps.short_name == namespace_component:
                    break
                root_namespace_perhaps = root_namespace_perhaps.get_nested_namespace(
                    namespace_component, create_if_missing=True
                )
            if root_namespace_perhaps == index:
                raise RuntimeError(f"Type {dsdl_type} has no namespace components and is therefore malformed.")
            root_namespace_perhaps.add_data_type(dsdl_type, input_types, extension)

    @classmethod
    def Identity(cls, output_path: Path, lctx: LanguageContext) -> "Namespace":
        """
        Create a namespace identity object. This is a namespace with no root directory and no parent. It can be used
        as a Namespace index object.

        :param Path output_path: The base path under which all namespaces and datatypes should be generated.
        :param LanguageContext lctx: The language context to use when building the namespace.
        :return: A namespace identity object.
        :rtype: Namespace
        """
        return cls("", output_path, lctx)

    @singledispatchmethod
    @classmethod
    def read_namespace(
        cls,
        index: "Namespace",
        root_namespace_directory: Path | str,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        allow_root_namespace_name_collision: bool = True,
    ) -> "Namespace":
        """
        Read a namespace from a root namespace directory.

        :param Namespace index: The index namespace to add the new namespaces and types to.
        :param Path | str root_namespace_directory: The root namespace directory.
        :param Path | str | Iterable[Path | str] lookup_directories: The directories to search for dsdl files.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :param bool allow_root_namespace_name_collision: Allow root namespace name collisions.
        :return: A new namespace index that contains trees of datatypes.
        """
        if not index.is_index:
            raise ValueError("Namespace passed in as index argument is not an index namespace.")

        composite_types = pydsdl.read_namespace(
            root_namespace_directory,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
            allow_root_namespace_name_collision,
        )
        Namespace.add_types(index, [(dsdl_type, []) for dsdl_type in composite_types])
        return index

    @read_namespace.register
    @classmethod
    def _(
        cls,
        output_path: str,
        lctx: LanguageContext,
        root_namespace_directory: Path | str,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        allow_root_namespace_name_collision: bool = True,
    ) -> pydsdl.Any:
        """
        Read a namespace from a root namespace directory.

        :param str output_path: The base path under which all namespaces and datatypes should be generated.
        :param LanguageContext lctx: The language context to use when building the namespace.
        :param Path | str root_namespace_directory: The root namespace directory.
        :param Path | str | Iterable[Path | str] lookup_directories: The directories to search for dsdl files.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :param bool allow_root_namespace_name_collision: Allow root namespace name collisions.
        :return: A new namespace index that contains trees of datatypes.
        """
        return cls.read_namespace(
            cls.Identity(Path(output_path), lctx),
            root_namespace_directory,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
            allow_root_namespace_name_collision,
        )

    @read_namespace.register
    @classmethod
    def _(
        cls,
        output_path: Path,
        lctx: LanguageContext,
        root_namespace_directory: Path | str,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        allow_root_namespace_name_collision: bool = True,
    ) -> pydsdl.Any:
        """
        Read a namespace from a root namespace directory.

        :param Path output_path: The base path under which all namespaces and datatypes should be generated.
        :param LanguageContext lctx: The language context to use when building the namespace.
        :param Path | str root_namespace_directory: The root namespace directory.
        :param Path | str | Iterable[Path | str] lookup_directories: The directories to search for dsdl files.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :param bool allow_root_namespace_name_collision: Allow root namespace name collisions.
        :return: A new namespace index that contains trees of datatypes.
        """
        return cls.read_namespace(
            cls.Identity(output_path, lctx),
            root_namespace_directory,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
            allow_root_namespace_name_collision,
        )

    @singledispatchmethod
    @classmethod
    def read_files(
        cls,
        index: "Namespace",
        dsdl_files: Path | str | Iterable[Path | str],
        root_namespace_directories_or_names: Path | str | Iterable[Path | str] | None,
        jobs: int = 0,
        job_timeout_seconds: float = 0,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        omit_dependencies: bool = False,
    ) -> "Namespace":
        """
        For a given set of dsdl_files, read the files and build a namespace tree.

        :param Namespace index: The index namespace to add the new namespaces and types to.
        :param Path | str | Iterable[Path | str] dsdl_files: The dsdl files to read.
        :param Path | str | Iterable[Path | str] root_namespace_directories_or_names: See :meth:`pydsdl.read_files`.
        :param int jobs: The number of parallel jobs to allow when reading multiple files. 0 Indicates no limit and 1
                   diasallows all parallelism.
        :param float job_timeout_seconds: Maximum time in fractional seconds any one read file job is allowed to take
                     before timing out. 0 disables timeouts.
        :param Path | str | Iterable[Path | str] lookup_directories: See :meth:`pydsdl.read_files`.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :return: A new namespace index that contains trees of datatypes.
        """
        if not index.is_index:
            raise ValueError("Namespace passed in as index argument is not an index namespace.")

        args = (
            root_namespace_directories_or_names,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
        )
        if jobs == 1:
            # Don't use multiprocessing when jobs is 1.
            return _read_files_strategy(index, NotAsyncResult, dsdl_files, job_timeout_seconds, omit_dependencies, args)
        else:
            with multiprocessing.pool.Pool(processes=None if jobs == 0 else jobs) as pool:
                return _read_files_strategy(
                    index, pool.apply_async, dsdl_files, job_timeout_seconds, omit_dependencies, args
                )

    @read_files.register
    @classmethod
    def _(
        cls,
        output_path: Path,
        lctx: LanguageContext,
        dsdl_files: Path | str | Iterable[Path | str] | None,
        root_namespace_directories_or_names: Path | str | Iterable[Path | str] | None,
        jobs: int = 0,
        job_timeout_seconds: float = 0,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        omit_dependencies: bool = False,
    ) -> pydsdl.Any:
        """
        For a given set of dsdl_files, read the files and build a namespace tree.

        :param Path output_path: The base path under which all namespaces and datatypes should be generated.
        :param LanguageContext lctx: The language context to use when building the namespace.
        :param Path | str | Iterable[Path | str] dsdl_files: The dsdl files to read.
        :param Path | str | Iterable[Path | str] root_namespace_directories_or_names: See :meth:`pydsdl.read_files`.
        :param int jobs: The number of parallel jobs to allow when reading multiple files. 0 Indicates no limit and 1
                   diasallows all parallelism.
        :param float job_timeout_seconds: Maximum time in fractional seconds any one read file job is allowed to take
                     before timing out. 0 disables timeouts.
        :param Path | str | Iterable[Path | str] lookup_directories: See :meth:`pydsdl.read_files`.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :return: A new namespace index that contains trees of datatypes.
        """
        return cls.read_files(
            Namespace.Identity(output_path, lctx),
            dsdl_files,
            root_namespace_directories_or_names,
            jobs,
            job_timeout_seconds,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
            omit_dependencies,
        )

    @read_files.register
    @classmethod
    def _(
        cls,
        output_path: str,
        lctx: LanguageContext,
        dsdl_files: Path | str | Iterable[Path | str] | None,
        root_namespace_directories_or_names: Path | str | Iterable[Path | str] | None,
        jobs: int = 0,
        job_timeout_seconds: float = 0,
        lookup_directories: Path | str | Iterable[Path | str] | None = None,
        print_output_handler: Callable[[Path, int, str], None] | None = None,
        allow_unregulated_fixed_port_id: bool = False,
        omit_dependencies: bool = False,
    ) -> pydsdl.Any:
        """
        For a given set of dsdl_files, read the files and build a namespace tree.

        :param str output_path: The base path under which all namespaces and datatypes should be generated.
        :param LanguageContext lctx: The language context to use when building the namespace.
        :param Path | str | Iterable[Path | str] dsdl_files: The dsdl files to read.
        :param Path | str | Iterable[Path | str] root_namespace_directories_or_names: See :meth:`pydsdl.read_files`.
        :param int The number of parallel jobs to allow when reading multiple files. 0 Indicates no limit and 1
                   diasallows all parallelism.
        :param float job_timeout_seconds: Maximum time in fractional seconds any one read file job is allowed to take
                     before timing out. 0 disables timeouts.
        :param Path | str | Iterable[Path | str] lookup_directories: See :meth:`pydsdl.read_files`.
        :param Callable[[Path, int, str], None] print_output_handler: A callback to handle print output.
        :param bool allow_unregulated_fixed_port_id: Allow unregulated fixed port ids.
        :return: A new namespace index that contains trees of datatypes.
        """
        return cls.read_files(
            Namespace.Identity(Path(output_path), lctx),
            dsdl_files,
            root_namespace_directories_or_names,
            jobs,
            job_timeout_seconds,
            lookup_directories,
            print_output_handler,
            allow_unregulated_fixed_port_id,
            omit_dependencies,
        )

    DefaultOutputStem = "_"

    def __init__(
        self,
        full_namespace: str,
        namespace_dir: Path,
        language_context: LanguageContext,
        parent: "Namespace | None" = None,
    ):
        if full_namespace.startswith("."):
            full_namespace = full_namespace[1:]

        self._language_context = language_context
        self._parent = parent

        if len(full_namespace) == 0:

            # Identity namespace
            if parent is not None:
                raise ValueError("Identity namespaces must not have a parent.")
            self._namespace_components: List[str] = []
            self._namespace_components_stropped: List[str] = []
            self._full_namespace = ""
            self._base_output_path = self._output_folder = namespace_dir
            self._source_folder = Path("")
            self._short_name = ""
        else:
            # Root or nested namespace
            if parent is None:
                raise ValueError("Non-identity namespaces must have a parent.")
            if len(namespace_dir.name) == 0:
                raise ValueError("Root namespace directory must have a name.")
            self._namespace_components, self._namespace_components_stropped = self.strop_namespace(
                full_namespace, language_context
            )
            if self._namespace_components[-1] != namespace_dir.name:
                raise ValueError(f"Namespace {full_namespace} does not match root namespace directory {namespace_dir}")
            self._base_output_path = parent.base_output_path
            self._full_namespace = ".".join(self._namespace_components_stropped)
            self._output_folder = Path(self._base_output_path / Path(*self._namespace_components_stropped))
            self._source_folder = namespace_dir / Path(*self._namespace_components[1:])
            self._short_name = self._namespace_components_stropped[-1]

        target_language = language_context.get_target_language()
        output_stem = target_language.get_config_value(Language.WKCV_NAMESPACE_FILE_STEM, self.DefaultOutputStem)
        output_path = self._output_folder / Path(output_stem)
        self._output_path = output_path.with_suffix(
            target_language.get_config_value(Language.WKCV_DEFINITION_FILE_EXTENSION)
        )
        self._data_type_to_outputs: dict[pydsdl.CompositeType, Generatable] = {}
        self._nested_namespaces: dict[str, Namespace] = {}

        if self._parent is not None and not self.is_index:  # pragma: no cover
            self._parent._nested_namespaces[self._namespace_components[-1]] = self

    # +--[PROPERTIES]-----------------------------------------------------------------------------------------------+
    @property
    def output_folder(self) -> Path:
        """
        The folder where this namespace's output file and datatypes are generated.
        """
        return self._output_folder

    @property
    def output_path(self) -> Path:
        """
        Path to the namespace's output file.
        """
        return self._output_path

    @property
    def parent(self) -> "Namespace | None":
        """
        The parent namespace of this namespace or None if this is a root namespace.
        """
        return self._parent

    @property
    def root_namespace(self) -> "Namespace":
        """
        The root namespace of this namespace.
        :raises RuntimeError: If this is an index namespace.
        """
        if self.is_index:
            raise RuntimeError("Index namespace has no root namespace.")

        root_maybe = self
        while not root_maybe.is_root:
            # All non-index namespaces have a parent and all non-index namespaces are either a root or have a root
            # ancestor.
            root_maybe = root_maybe.parent  # type: ignore

        return root_maybe

    @property
    def base_output_path(self) -> Path:
        """
        The folder under which artifacts are generated.
        """
        return self._base_output_path

    @property
    def is_index(self) -> bool:
        """
        True if this namespace is an index namespace.
        """
        return self._parent is None

    @property
    def is_root(self) -> bool:
        """
        True if this namespace is a root namespace.
        """
        return len(self._namespace_components) == 1

    @property
    def is_nested(self) -> bool:
        """
        True if this namespace is a nested namespace.
        """
        return len(self._namespace_components) > 1

    # +--[PUBLIC]--------------------------------------------------------------------------------------------------+
    def get_index_namespace(self) -> "Namespace":
        """
        The index namespace is a meta-namespace that is empty and has no data types. It contains
        the root output folder, a common language context, and all the namespaces in a tree of DSDL types. It is used to
        generate index files that reference all the generated files in the tree at once. Logically, it is the root of
        the tree and is always the parent of each root namespace. The taxonomy of namespaces is therefore ::

                                          ┌────────────────┐
                                          │  CompoundType  │
                                          │   «duck type»  │
                                          └────────────────┘
                                           ▲             ▲
                        ┌──────────────────┘             │
                        │                                │
                        │              ┌──────┐          │
                        │              │ Path │          │
                        │              │      │          │
                        │              └──────┘          │
                        │               ▲    ▲           │
                        │               │    │           │
                  ┌─────┴─────┐ ┌───────┴┐  ┌┴───────┐   │
                  │ Namespace │ │ Folder │  │  File  │   │
                  │           │ │        │  │        │   │
                  └───────────┘ └────────┘  └────────┘   │
                   ▲    ▲   ▲     ^    ^           ^     │
                   │    │   │     :    :.......    :     │
                   │    │   │     :           :    :     │
                   │    │   └──── : ─────┐    :    :     │
                   │    │         :      │    :    :     │
                   │    └─────┐   :      │    :    :     │
                   │          │   :      │    :    :     │
                ┌──┴──────┐ ┌─┴───┴──┐ ┌─┴────┴─┐ ┌┴─────┴────┐
                │  index  │ │  root  │ │ nested │ │ DSDL Type │
                │         │ │        │ │        │ │           │
                └─────────┘ └────────┘ └────────┘ └───────────┘

        :return: The index namespace.

        .. invisible-code-block: python

            from nunavut._namespace import Namespace
            from nunavut.lang import LanguageContext, LanguageContextBuilder
            from pathlib import Path

            lctx = (
                LanguageContextBuilder()
                    .set_target_language("c")
                    .create()
            )

            base_path = gen_paths.out_dir
            root_namespace_dir = base_path / Path("uavcan")
            root_namespace_dir.mkdir(exist_ok=True)
            nested_namespace_dir = root_namespace_dir / Path("node")
            nested_namespace_dir.mkdir(exist_ok=True)

        .. code-block:: python

            # This is the index namespace identity.
            index_ns = Namespace.Identity(base_path, lctx)
            ns = Namespace("uavcan", root_namespace_dir, lctx, index_ns)

            # This is a root namespace identity.
            assert ns.get_index_namespace() == index_ns

        """
        namespace = self
        while not namespace.is_index:
            # parent is always set for non-index namespaces
            namespace = namespace.parent  # type: ignore
        return namespace

    def get_language_context(self) -> LanguageContext:
        """
        The generated software language context the namespace is within.
        """
        return self._language_context

    def get_root_namespace(self, root_namespace_name: str, create_if_missing: bool = False) -> "Namespace":
        """
        Retrieves or creates a root namespace object for a given name.

        :param Path root_namespace_name: The root namespace name to get.
        :param bool create_if_missing: If True, the namespace will be created if it does not exist.
        :return: The root namespace object.
        :raises KeyError: If the namespace was not found and create_if_missing is False.
        """
        index = self.get_index_namespace()
        return index.get_nested_namespace(root_namespace_name, create_if_missing=create_if_missing)

    def get_nested_namespaces(self) -> Iterator["Namespace"]:
        """
        Get an iterator over all the nested namespaces within this namespace.
        This is a shallow iterator that only provides directly nested namespaces.
        """
        return iter(self._nested_namespaces.values())

    def get_nested_namespace(self, namespace_name: str, create_if_missing: bool = False) -> "Namespace":
        """
        Get a nested namespace by name. Note, this is not recursive.

        :param str namespace_name: The name of the nested namespace to get.
        :param bool create_if_missing: If True, the namespace will be created if it does not exist.
        :return: The nested namespace.
        :raises KeyError: If the namespace was not found and create_if_missing is False.
        """
        try:
            return self._nested_namespaces[namespace_name]
        except KeyError as e:
            if not create_if_missing:
                raise e

        return Namespace(
            ".".join(self._namespace_components + [namespace_name]),
            self.source_file_path / Path(namespace_name),
            self._language_context,
            self,
        )

    def get_nested_types(self) -> ItemsView[pydsdl.CompositeType, Generatable]:
        """
        Get a view of a tuple relating datatypes in this namespace to the path for the
        type's generated output. This is a shallow view including only the types
        directly within this namespace.
        """
        return self._data_type_to_outputs.items()

    def get_all_datatypes(self) -> Generator[Tuple[pydsdl.CompositeType, Generatable], None, None]:
        """
        Generates tuples relating datatypes at and below this namespace to the path
        for each type's generated output.
        """
        yield from self._recursive_data_type_generator(self)

    def get_all_namespaces(self) -> Generator[Tuple["Namespace", Path], None, None]:
        """
        Generates tuples relating nested namespaces at and below this namespace to the path
        for each namespace's generated output.
        """
        yield from self._recursive_namespace_generator(self)

    def get_all_types(self) -> Generator[Tuple[pydsdl.Any, Generatable | Path], None, None]:
        """
        Generates tuples relating datatypes and nested namespaces at and below this
        namespace to the path for each type's generated output.
        """
        yield from self._recursive_data_type_and_namespace_generator(self)

    def find_output_path_for_type(self, compound_type: "Namespace | pydsdl.CompositeType") -> Path:
        """
        Searches the entire namespace tree to find a mapping of the type to an
        output file path.

        :param pydsdl.CompositeType compound_type: A Namespace or pydsdl.CompositeType to find the output pathfor.
        :return: The path where a file will be generated for a given type.
        :raises KeyError: If the type was not found in this namespace tree.
        """
        if isinstance(compound_type, Namespace):
            return compound_type.output_path
        else:
            # pylint: disable=protected-access
            root_namespace = self.get_index_namespace().get_root_namespace(compound_type.source_file_path_to_root.name)
            return root_namespace._bfs_search_for_output_path(compound_type)  # pylint: disable=protected-access

    def add_data_type(
        self, dsdl_type: pydsdl.CompositeType, input_types: List[pydsdl.CompositeType], extension: str | None
    ) -> Generatable:
        """
        Add a datatype to this namespace.

        :param pydsdl.CompositeType dsdl_type: The datatype to add.
        :param str extension: The file extension to use for the generated file. If None, the
                                extension will be determined by the target language.
        :return: A path to the file this type will be generated in.
        """

        if self.is_index:
            raise RuntimeError("Cannot add types to an index namespace.")

        _, stropped_ns = self.strop_namespace(dsdl_type.full_namespace, self._language_context)
        if stropped_ns[-1] != self._short_name:
            raise ValueError(
                f"Type {dsdl_type.full_name} does not belong in namespace {self._full_namespace}. "
                f"Expected namespace {self._full_namespace}, got {dsdl_type.full_namespace}."
            )

        language = self._language_context.get_target_language()
        if extension is None:
            extension = language.get_config_value(Language.WKCV_DEFINITION_FILE_EXTENSION)
        output_file = Path(self._base_output_path) / IncludeGenerator.make_path(dsdl_type, language, extension)
        output_generatable = Generatable(dsdl_type, input_types, output_file)
        self._data_type_to_outputs[dsdl_type] = output_generatable
        return output_generatable

    # +--[DUCK TYPING: pydsdl.CompositeType]-----------------------------------------------------------------------+
    @property
    def short_name(self) -> str:
        """
        See :py:attr:`pydsdl.CompositeType.short_name`
        """
        return self._short_name

    @property
    def full_name(self) -> str:
        """
        See :py:attr:`pydsdl.CompositeType.full_name`
        """
        return self._full_namespace

    @property
    def full_namespace(self) -> str:
        """
        See :py:attr:`pydsdl.CompositeType.full_namespace`
        """
        return self._full_namespace

    @property
    def namespace_components(self) -> List[str]:
        """
        See :py:attr:`pydsdl.CompositeType.namespace_components`
        """
        return self._namespace_components

    @property
    def source_file_path(self) -> Path:
        """
        See :py:attr:`pydsdl.CompositeType.source_file_path`
        Note that, for Namespace objects, this path is always relative since a single namespace may contain sources
        from files rooted in different directory trees. For example, a namespace may a type
        "/path/to/animals/mammals/Dog.1.0.dsdl" and another type "/a/different/path/to/animals/mammals/Cat.1.0.dsdl".
        """
        return self._source_folder

    @property
    def source_file_path_to_root(self) -> Path:
        """
        See :py:attr:`pydsdl.CompositeType.source_file_path_to_root`
        Note that, for Namespace objects, this path is always relative. See :py:attr:`source_file_path` for more.
        """
        return self.root_namespace.source_file_path

    @property
    def data_types(self) -> KeysView[pydsdl.CompositeType]:
        """
        See :py:attr:`pydsdl.CompositeType.data_types`
        """
        return self._data_type_to_outputs.keys()

    @property
    def attributes(self) -> List[pydsdl.CompositeType]:
        """
        See :py:attr:`pydsdl.CompositeType.attributes`
        """
        return []

    # +--[PYTHON DATA MODEL]--------------------------------------------------------------------------------------+

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Namespace):
            return self._full_namespace == other._full_namespace
        else:
            return False

    def __str__(self) -> str:
        return self.full_name

    def __hash__(self) -> int:
        return hash(self._full_namespace)

    # +--[PRIVATE]------------------------------------------------------------------------------------------------+

    def _bfs_search_for_output_path(self, data_type: pydsdl.CompositeType) -> Path:
        search_queue: Deque[Namespace] = collections.deque()
        search_queue.appendleft(self)
        while len(search_queue) > 0:
            namespace = search_queue.pop()
            try:
                return namespace._data_type_to_outputs[data_type]  # pylint: disable=protected-access
            except KeyError:
                pass
            for nested_namespace in namespace._nested_namespaces.values():  # pylint: disable=protected-access
                search_queue.appendleft(nested_namespace)

        raise KeyError(data_type)

    @classmethod
    def _recursive_data_type_generator(
        cls, namespace: "Namespace"
    ) -> Generator[Tuple[pydsdl.CompositeType, Generatable], None, None]:
        for data_type, output_path in namespace.get_nested_types():
            yield (data_type, output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_data_type_generator(nested_namespace)

    @classmethod
    def _recursive_namespace_generator(cls, namespace: "Namespace") -> Generator[Tuple["Namespace", Path], None, None]:
        yield (namespace, namespace.output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_namespace_generator(nested_namespace)

    @classmethod
    def _recursive_data_type_and_namespace_generator(
        cls, namespace: "Namespace"
    ) -> Generator[Tuple[pydsdl.Any, Path | Generatable], None, None]:
        yield (namespace, namespace.output_path)

        for data_type, output_path in namespace.get_nested_types():
            yield (data_type, output_path)

        for nested_namespace in namespace.get_nested_namespaces():
            yield from cls._recursive_data_type_and_namespace_generator(nested_namespace)


# +---------------------------------------------------------------------------+


def build_namespace_tree(
    types: List[pydsdl.CompositeType],
    root_namespace_dir: str | Path,
    output_dir: str | Path,
    language_context: LanguageContext,
) -> Namespace:
    """
    Generates a :class:`nunavut.Namespace` tree.

    .. note::

        Deprecated. Use :method:`Namespace.add_types` instead. build_namespace_tree creates a new a :class:`Namespace`
        index internally which may lead to unexpected behavior if calling this method multiple times. Furthermore, it
        cannot associate output files with their dependent types and is ambiguous about the root namespace directory.

    Given a list of pydsdl types, this method returns a root :class:`nunavut.Namespace`.
    The root :class:`nunavut.Namespace` is the top of a tree where each node contains
    references to nested :class:`nunavut.Namespace` and to any :code:`pydsdl.CompositeType`
    instances contained within the namespace.

    :param list types: A list of pydsdl types.
    :param str | Path root_namespace_dir: The root namespace directory. This is the directory that contains the
            namespaces's datatypes and nested namespaces.

        .. note::
            Root namespace directories are determined by the source file path of individual types so it is possible
            to pass in a list of types that are not available in the returned Namespace. Only types that are within
            this root namespace directory will be included in the returned Namespace.

    :param str | Path output_dir: The base directory under which all generated files will be created.
    :param nunavut.LanguageContext language_context: The language context to use when building
            :class:`nunavut.Namespace` objects.
    :return: The root :class:`nunavut.Namespace`.
    :rtype: nunavut.Namespace

    """

    index = Namespace.Identity(Path(output_dir), language_context)
    Namespace.add_types(index, [(dsdl_type, []) for dsdl_type in types])
    return index.get_root_namespace(Path(root_namespace_dir).name, create_if_missing=True)


# +---------------------------------------------------------------------------+



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""

.. autodata:: __version__
"""

import sys

__version__ = "3.0.1b1"  # please update NunavutConfigVersion.cmake if changing the major or minor version.
__license__ = "MIT"
__author__ = "OpenCyphal"
__copyright__ = (
    "Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved. Copyright (c) OpenCyphal "
    "Development Team <opencyphal.org>."
)
__email__ = "maintainers@opencyphal.org"
__pydsdl_version__ = ">= 1.24.3"


def _parse_version_string(tag: str, pattern: str, major_minor_only: bool) -> list:  # pragma: no cover
    """
    Given a version string, extract the version number as a list of strings.
    """
    from re import match as rematch  # pylint: disable=import-outside-toplevel

    if tag.startswith("refs/tags/"):
        tag = tag.split("/")[-1]
    match = rematch(pattern, tag)
    if match is None:
        groups = ["0", "0", "0"]
    else:
        groups = match.group(1).split(".")

    if major_minor_only and len(groups) > 1:
        return groups[0:2]
    else:
        return groups


def _fail_on_mismatch(tag: str, tag_triplet_pattern: str, major_minor_only: bool) -> None:  # pragma: no cover
    """
    Call exit if the git tag does not match the embedded version.
    """
    version_from_tag = _parse_version_string(tag, tag_triplet_pattern, major_minor_only)
    nunavut_version = _parse_version_string(__version__, tag_triplet_pattern, major_minor_only)
    if ".".join(version_from_tag) != ".".join(nunavut_version):
        sys.stderr.write(
            f"tagged version {'.'.join(version_from_tag)} does not match the embedded version {__version__}.\r\n"
        )
        sys.exit(-1)


if __name__ == "__main__":  # pragma: no cover
    from argparse import ArgumentParser  # pylint: disable=import-outside-toplevel

    parser = ArgumentParser(description="Get the version of the package.")
    parser.add_argument("--verbose", "-v", action="store_true", help="Print more information.")
    parser.add_argument("--git", action="store_true", help="Get the version from the latest git tag.")
    parser.add_argument("--git-branch", default="main", help="The git branch to use.")
    parser.add_argument("--tag-match", default="v*", help="The git tag pattern to match.")
    parser.add_argument("--tag", help="A manually provided git tag. Overrides --git.")
    parser.add_argument("--major-minor-only", action="store_true", help="Only use major and minor version.")
    parser.add_argument(
        "--tag-triplet-pattern",
        default=r"^v?(\d+\.\d+(?:\.\d+)?(?:\.[a-zA-Z]\w*|[a-zA-Z]\w*)?)$",
        help="A regex pattern to extract the version from the git tag (supports PEP 440).",
    )
    parser.add_argument(
        "--fail-on-mismatch", action="store_true", help="Fail if the git tag does not match the embedded version."
    )
    args = parser.parse_args()

    local_version = ".".join(_parse_version_string(__version__, args.tag_triplet_pattern, args.major_minor_only))
    if args.verbose:
        print(f"nunavut version: {local_version}")
    else:
        print(local_version, end="")

    if args.tag is not None:
        if args.verbose:
            print(f"Git-tag version from command-line is {args.tag}")
        if args.fail_on_mismatch:
            _fail_on_mismatch(args.tag, args.tag_triplet_pattern, args.major_minor_only)
    elif args.git:
        from subprocess import run as run_subprocess  # pylint: disable=import-outside-toplevel

        # amazonq-ignore-next-line
        completed = run_subprocess(
            ["git", "describe", args.git_branch, "--tags", "--abbrev=0", f"--match={args.tag_match}"],
            capture_output=True,
            check=False,
        )
        if completed.returncode == 0:
            git_tag_version = completed.stdout.decode("utf-8").strip()

            if args.verbose:
                print(f"Git-tagged version on {args.git_branch} is {git_tag_version}")
            if args.fail_on_mismatch:
                _fail_on_mismatch(git_tag_version, args.tag_triplet_pattern, args.major_minor_only)
        elif args.fail_on_mismatch:
            if args.verbose:
                sys.stderr.write(f"Failed to get git tag: {completed.stderr.decode('utf-8')}\r\n")
            sys.exit(completed.returncode)
        elif args.verbose:
            sys.stdout.write(f"Failed to get git tag: {completed.stderr.decode('utf-8')}\r\n")



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Command-line for using nunavut and jinja to generate code
from dsdl definitions.
"""

import argparse
import sys
import textwrap
from pathlib import Path
from typing import Any, Optional, Tuple, Type, TypeVar, cast


class _LazyVersionAction(argparse._VersionAction):
    """
    Changes argparse._VersionAction so we only load nunavut.version
    if the --version action is requested.
    """

    # pylint: disable=protected-access

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: Optional[str] = None,
    ) -> None:
        # pylint: disable=import-outside-toplevel
        from nunavut._version import __version__

        parser._print_message(__version__, sys.stdout)
        parser.exit()


def keyvalue(arg: str) -> Tuple[str, Any]:
    """
    Parse a key=value string into a tuple.

    :param arg: String in the format "key=value"
    :return: Tuple of (key, value)
    :raises ValueError: If the string is not in key=value format
    """
    try:
        key, value = arg.split("=", 1)
        value_stripped: Any = value.strip()
        if value_stripped == "false":
            value_stripped = False
        elif value_stripped == "true":
            value_stripped = True
        return (key.strip(), value_stripped)
    except ValueError:
        raise ValueError(f'"{arg}" is not in key=value format') from None


ParserT = TypeVar("ParserT")
"""
Type variable for the concrete parser to create.
"""


def _make_parser(parser_type: Type[ParserT]) -> ParserT:
    """
    Defines the command-line interface. Provided as a separate factory method to
    support sphinx-argparse documentation.

    :param parser_type: The type of parser to create. This should be a subclass of argparse.ArgumentParser
    :return: A parser instance with the command-line interface defined.
    :raises ValueError: If parser_type is not a subclass of argparse

    .. invisible-code-block: python

        from nunavut.cli import _make_parser
        from pytest import raises

        with raises(ValueError):
            _make_parser(str)

    """

    epilog = textwrap.dedent("""

        Copyright (C) OpenCyphal Development Team  <opencyphal.org>
        Copyright Amazon.com Inc. or its affiliates.
        Released under SPDX-License-Identifier: MIT

        **Example Usage** ::

            # This would include j2 templates for a folder named 'c_jinja'
            # and generate .h files into a directory named 'include' for
            # the uavcan.node.Heartbeat.1.0 data type and its dependencies

            nnvg --outdir include --templates c_jinja -e .h dsdl/uavcan:node/7509.Heartbeat.1.0.dsdl

        ᓄᓇᕗᑦ
    """)

    if not issubclass(parser_type, argparse.ArgumentParser):
        raise ValueError("parser_type must be a subclass of argparse.ArgumentParser")

    parser = parser_type(
        description="Generate code from Cyphal DSDL using pydsdl and jinja2",
        epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "target_files_or_root_namespace",
        nargs="*",
        help=textwrap.dedent("""

        One or more dsdl files to generate from.

        All dependent types found in the target dsdl files will be generated and must be available
        either as another target or as a dsdl file under one of the --lookup-dir directories.

        A --lookup-dir argument for each unique root path among the list of files is required if
        not using the colon syntax.

        Colon Syntax
        ------------

            The standard syntax allows the path to the root to be specified at the same
            time as the type ::

                path/to/root:name/space/Type.1.0.dsdl

            This also adds the path to a list of valid paths. You can continue to specify
            it (duplicates are ignored) or you can specify it once ::

                path/to/root:name/space/Type.1.0.dsdl name/space/Type.1.0.dsdl

                ...is the same as...

                path/to/root:name/space/Type.1.0.dsdl path/to/root:name/space/Type.1.0.dsdl

        Globular Expansion
        ------------------

            Any target paths that are folders will be considered to be root namespaces. nnvg
            will then search for all dsdl files under that folder and generate target paths
            for all .dsdl files found. To disable this behavior use the --no-target-namespaces
            argument which will cause an error if a folder is provided as a target.

    """).lstrip(),
    )

    parser.add_argument(
        "--no-target-namespaces",
        action="store_true",
        help=textwrap.dedent("""

        If provided then all target paths must be to individual DSDL files and not folders. If
        set and a folder is provided as a target an error will be raised.

        Normally, if a folder is provided as a target path, nnvg will search for all dsdl files
        under that folder and generate target paths for all .dsdl files found using the
        path itself as the root namespace.

        see target_files_or_root_namespace for more information.
    """).lstrip(),
    )

    parser.add_argument(
        "--lookup-dir",
        "-I",
        action="append",
        help=textwrap.dedent("""

        List of other namespace directories containing data type definitions that are referred to
        from the target root namespace. For example, if you are reading a vendor-specific namespace,
        the list of lookup directories should always include a path to the standard root namespace
        "uavcan", otherwise the types defined in the vendor-specific namespace won't be able to use
        data types from the standard namespace.

        For a given target set of dsdl files this argument is required to specify a set of valid
        paths to or folder names of root namespaces. For example ::

            nnvg types/animals/felines/Tabby.1.0.dsdl types/animals/canines/Boxer.1.0.dsdl

        will fail unless the path describing the root is provided ::

            nnvg --lookup-dir types/animals types/animals/cats/Tabby.1.0.dsdl \\
                                            types/animals/dogs/Boxer.1.0.dsdl

        If multiple roots are targeted then each root path will need to be enabled ::

            nnvg -I types/animals -I types/plants types/animals/cats/Tabby.1.0.dsdl \\
                                                  types/plants/trees/Fir.1.0.dsdl


        For target files this argument is required to specify a set of valid paths to or folder
        names of root namespaces, however; an additional syntax is supported where the root for a
        target file can be specified as part of the target path using a colon to separate the two ::

            nnvg types/animals:cats/Tabby.1.0.dsdl types/plants:trees/Fir.1.0.dsdl


        This is the recommended syntax as it allows all target files to be specified as relative
        paths and reserves this argument to specify concrete lookup directories ::

            nnvg --lookup-dir /path/to/types types/animals:mammals/cats/Tabby.1.0.dsdl \\
                                             types/plants:trees/conifers/Fir.1.0.dsdl

        Additional directories can also be specified through an environment variable
        DSDL_INCLUDE_PATH where the path entries are separated by colons ":" on posix systems and
        ";" on Windows ::

            DSDL_INCLUDE_PATH=/path/to/types/animals:/path/to/types/plants \\
                nnvg animals:mammals/cats/Tabby.1.0.dsdl \\
                     plants:trees/conifers/Fir.1.0.dsdl

        CYPHAL_PATH is similarly used (and formatted) but the paths in this variable are listed and
        each folder under each path will be a lookup directory ::

            CYPHAL_PATH=/path/to/types nnvg animals:mammals/cats/Tabby.1.0.dsdl \\
                                            plants:trees/conifers/Fir.1.0.dsdl

    """).lstrip(),
    )

    parser.add_argument("--outdir", "-O", default="nunavut_out", help="output directory")

    parser.add_argument(
        "--target-language",
        "-l",
        help=textwrap.dedent("""

        Language support to install into the templates.

        If provided then the output extension (-e) can be inferred otherwise the output
        extension must be provided.

    """).lstrip(),
    )

    # +-----------------------------------------------------------------------+
    # | Extended Generation Options
    # +-----------------------------------------------------------------------+

    extended_group = parser.add_argument_group(
        "extended options",
        description=textwrap.dedent("""

        Additional options to control output generation.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--templates-dir",
        "--templates",
        type=Path,
        help=textwrap.dedent("""

        Paths to a directory containing templates to use when generating code.

        Templates found under these paths will override the built-in templates for a
        given language.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--support-templates-dir",
        "--support-templates",
        type=Path,
        help=textwrap.dedent("""

        Paths to a directory containing templates to use when generating support code.

        Templates found under these paths will override the built-in support templates for a
        given language.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--fallback-to-builtin-templates",
        "-tfb",
        action="store_true",
        help=textwrap.dedent("""

        Normally, if providing a custom templates directory, the built-in templates are not
        searched for. This option will cause the built-in templates to be searched for if a
        template is not found in the custom templates directory first.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--index-file",
        type=Path,
        action="append",
        help=textwrap.dedent("""

        Index-files are generated from special templates that are given access to all types in
        all namespaces. This is useful for generating files that are not specific to a single
        type like dependency files, manifests, or aggregate headers.

        The template lookup paths are the same as for DSDL types but the template is selected
        based on the filename of the all-file first and falls back to `index.j2` if no specific
        template is found. Note that the DSDL type hierarchy will be searched first so you
        cannot name your index-file the same as a DSDL data type like `StructureType.j2`.

        Example ::

            # looks for all_types.j2 in the c_jinja template directory and generates
            # generated/include/all_types.h from all types in all DSDL namespaces.
            nnvg --index-file all_types.h --outdir generated/include --templates c_jinja \\
                path/to/types/animal:cat.1.0.dsdl \\
                path/to/types/animal:dog.1.0.dsdl

            # looks for manifest.j2 in the json_jinja template directory and generates
            # generated/include/manifest.json from all types in all DSDL namespaces.
            nnvg --index-file include/manifest.json --outdir generated --templates json_jinja \\
                path/to/types/animal:cat.1.0.dsdl \\
                path/to/types/animal:dog.1.0.dsdl

    """).lstrip(),
    )

    extended_group.add_argument(
        "--include-experimental-languages",
        "--experimental-languages",
        "-Xlang",
        action="store_true",
        help=textwrap.dedent("""

        Activate languages with unstable, experimental support.

        By default, target languages where support is not finalized are not
        enabled when running nunavut, to make it clear that the code output
        may change in a non-backwards-compatible way in future versions, or
        that it might not even work yet.

    """).lstrip(),
    )

    def extension_type(raw_arg: str) -> str:
        if raw_arg and not raw_arg.startswith("."):
            return "." + raw_arg
        else:
            return raw_arg

    extended_group.add_argument(
        "--output-extension",
        "-e",
        type=extension_type,
        help=textwrap.dedent("""
        The output extension for generated files. If target language is provided an extension is
        inferred based on language configuration. This option allows overriding this inference.

        Note that, if --target-language is omitted and this argument is provided the program
        will attempt to infer the language based on the file extension. This may lead to unexpected
        results (e.g. `--output-extension .h` generating C instead of C++).

        If a dot (.) is omitted one will be added, therefore; `-e h` and `-e .h` will both result
        in an extension of `.h`.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--generate-support",
        choices=["always", "never", "as-needed", "only"],
        default="as-needed",
        help=textwrap.dedent("""
        Change the criteria used to enable or disable support code generation.

        as-needed (default) - generate support if it is needed.
        always - always generate support code.
        never - never generate support code.
        only - only generate support code.

        Note that serialization logic is only one type of support code. This option covers all
        types of support code. Where `--omit-serialization-support` is set different types of
        support code may still be generated unless this option is set to `never`.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--generate-namespace-types",
        action="store_true",
        help=textwrap.dedent("""
        If enabled this script will generate source for namespaces.
        All namespaces including and under the root namespace will be treated as a
        pseudo-type and the appropriate template will be used. The generator will
        first look for a template with the stem "Namespace" and will then use the
        "Any" template if that is available. The name of the output file will be
        the default value for the --namespace-output-stem argument and can be
        changed using that argument.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--omit-serialization-support",
        "-pod",
        action="store_true",
        help=textwrap.dedent("""
        If provided then the types generated will be POD datatypes with no additional logic.
        By default types generated include serialization routines and additional support libraries,
        headers, or methods as needed. These additional support artifacts can be suppressed using
        the `--generate-support` option.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--namespace-output-stem",
        help="The name of the file generated when --generate-namespace-types is provided.",
    )

    extended_group.add_argument(
        "--no-overwrite",
        action="store_true",
        help=textwrap.dedent("""

        By default, generated files will be silently overwritten by
        subsequent invocations of the generator. If this argument is specified an
        error will be raised instead preventing overwrites.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--file-mode",
        default=0o444,
        type=lambda value: int(value, 0),
        help=textwrap.dedent("""

        The file-mode each generated file is set to after it is created.
        Note that this value is interpreted using python auto base detection.
        Because of this, to provide an octal value, you'll need to prefix your
        literal with '0o' (e.g. --file-mode 0o664).

    """).lstrip(),
    )

    extended_group.add_argument(
        "--allow-unregulated-fixed-port-id",
        action="store_true",
        help=textwrap.dedent("""

        Do not reject unregulated fixed port identifiers.
        This is a dangerous feature that must not be used unless you understand the
        risks. The background information is provided in the Cyphal specification.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--embed-auditing-info",
        action="store_true",
        help=textwrap.dedent("""

        If set, generators are instructed to add additional information in the form of
        language-specific comments or meta-data to use when auditing source code generated by
        Nunavut. This data may change based on the environment in use which may interfere with
        the reproducibility of your builds. For example, paths to input files used to generate
        a type may be included with this option where these paths will be different depending
        on the server used to run nnvg.

    """).lstrip(),
    )

    extended_group.add_argument(
        "--omit-dependencies",
        action="store_true",
        help=textwrap.dedent("""

        Disables the generation of dependent types. This is useful when setting up build
        rules for a project where the dependent types are generated separately.

    """).lstrip(),
    )

    # +-----------------------------------------------------------------------+
    # | Operation Options
    # +-----------------------------------------------------------------------+

    run_mode_group = parser.add_argument_group(
        "run mode options",
        description=textwrap.dedent("""

        Options that control the operation mode of the script.

    """).lstrip(),
    )

    run_mode_group.add_argument("--verbose", "-v", action="count", help="verbosity level (-v, -vv)")

    run_mode_group.add_argument("--version", action=_LazyVersionAction)

    run_mode_group.add_argument("--dry-run", "-d", action="store_true", help="If True then no files will be generated.")

    run_mode_group.add_argument(
        "--jobs",
        "-j",
        type=int,
        default=0,
        help=textwrap.dedent("""

        Limits the number of subprocesses nnvg can use to parallelize type discovery
        and code generation.

        If set to 0 then the number of jobs will be set to the number of CPUs available
        on the system.

        If set to 1 then no subprocesses will be used and all work will be done in th
        main process.

    """).lstrip(),
    )

    run_mode_group.add_argument(
        "--list-outputs",
        action="store_true",
        help=textwrap.dedent("""
        Emit a semicolon-separated list of files.
        (implies --dry-run)
        Emits files that would be generated if invoked without --dry-run.
        This command is useful for integrating with CMake and other build
        systems that need a list of targets to determine if a rebuild is
        necessary.

        If used with --list-inputs the list of inputs will be emitted first followed
        by the list of outputs. A single empty value will separate the two lists when
        using value-delimited formats. Use --list-format to control the output format
        including using json to avoid the need for an empty-value delimiter.

    """).lstrip(),
    )

    run_mode_group.add_argument(
        "--list-inputs",
        action="store_true",
        help=textwrap.dedent("""

        Emit a semicolon-separated list of files.
        (implies --dry-run)
        A list of files that are resolved given input arguments like templates.
        This command is useful for integrating with CMake and other build systems
        that need a list of inputs to determine if a rebuild is necessary.

        If used with --list-outputs the list of inputs will be emitted first followed
        by the list of outputs. A single empty value will separate the two lists. Use
        --list-format to control the output format including using json to avoid the
        need for an empty-value delimiter.

    """).lstrip(),
    )

    run_mode_group.add_argument(
        "--list-configuration",
        "-lc",
        action="store_true",
        help=textwrap.dedent("""

        Lists all configuration values resolved for the given arguments. Unlike --list-inputs
        and --list-outputs this command does *not* imply --dry-run but can be used in conjunction
        with it.

        This option is only available if --list-format is set to json.

    """).lstrip(),
    )

    run_mode_group.add_argument(
        "--list-format",
        default="scsv",
        choices=["csv", "scsv", "json", "json-pretty"],
        help=textwrap.dedent("""

        For commands that emit lists of files this option controls the format of the output.

        csv         - comma separated values
        scsv        - semicolon separated values
        json        - json formatted results
        json-pretty - json formatted results with indentation

    """).lstrip(),
    )

    run_mode_group.add_argument(
        "--list-to-file",
        type=Path,
        help=textwrap.dedent("""

        If provided then the output of --list-outputs, --list-inputs, or --list-configuration
        will also be written to the file specified. If the file exists it will be overwritten.
        This utf-8-encoded file will be written in the format specified by --list-format even
        if --dry-run is set.

    """).lstrip(),
    )

    # +-----------------------------------------------------------------------+
    # | Post-Processing Options
    # +-----------------------------------------------------------------------+

    ln_pp_group = parser.add_argument_group(
        "post-processing options",
        description=textwrap.dedent("""

        This options are all deprecated and will be removed in a future release.

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "--trim-blocks",
        action="store_true",
        help=textwrap.dedent("""

        If this is set to True the first newline after a block in a template
        is removed (block, not variable tag!).

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "--lstrip-blocks",
        action="store_true",
        help=textwrap.dedent("""

        If this is set to True leading spaces and tabs are stripped from the
        start of a line to a block in templates.

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "--pp-max-emptylines",
        type=int,
        help=textwrap.dedent("""

        If provided this will suppress generation of additional consecutive
        empty lines beyond the limit set by this argument.

        Note that this will insert a line post-processor which may reduce
        performance. Consider using a code formatter on the generated output
        to enforce whitespace rules instead.

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "--pp-trim-trailing-whitespace",
        action="store_true",
        help=textwrap.dedent("""

        Enables a line post-processor that will elide all whitespace at the
        end of each line.

        Note that this will insert a line post-processor which may reduce
        performance. Consider using a code formatter on the generated output
        to enforce whitespace rules instead.

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "-pp-rp",
        "--pp-run-program",
        help=textwrap.dedent("""

        Runs a program after each file is generated but before the file is
        set to read-only.

        example ::

            # invokes clang-format with the "in-place" argument on each file after it is
            # generated.

            nnvg --outdir include --templates c_jinja -e .h -pp-rp clang-format -pp-rpa=-i dsdl

    """).lstrip(),
    )

    ln_pp_group.add_argument(
        "-pp-rpa",
        "--pp-run-program-arg",
        action="append",
        help=textwrap.dedent("""

        Additional arguments to provide to the program specified by --pp-run-program.
        The last argument will always be the path to the generated file.

    """).lstrip(),
    )

    # +-----------------------------------------------------------------------+
    # | Language Options
    # +-----------------------------------------------------------------------+
    ln_opt_group = parser.add_argument_group(
        "language options",
        description=textwrap.dedent("""

        Options passed through to templates as `options` on the target language.

        Note that these arguments are passed though without validation, have no effect on the
        Nunavut library, and may or may not be appropriate based on the target language and
        generator templates in use.
    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--target-endianness",
        choices=["any", "big", "little"],
        help=textwrap.dedent("""

        Specify the endianness of the target hardware. This allows serialization
        logic to be optimized for different CPU architectures.

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--omit-float-serialization-support",
        action="store_true",
        help=textwrap.dedent("""

        Instruct support header generators to omit support for floating point operations
        in serialization routines. This will result in errors if floating point types are used,
        however; if you are working on a platform without IEEE754 support and do not use floating
        point types in your message definitions this option will avoid dead code or compiler
        errors in generated serialization logic.

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--enable-serialization-asserts",
        action="store_true",
        help=textwrap.dedent("""

        Instruct support header generators to generate language-specific assert statements as part
        of serialization routines. By default the serialization logic generated may make assumptions
        based on documented requirements for calling logic that could expose a system to undefined
        behavior. The alternative, for languages that do not support exception handling, is to
        use assertions designed to halt a program rather than execute undefined logic.

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--enable-override-variable-array-capacity",
        action="store_true",
        help=textwrap.dedent("""

        Instruct support header generators to add the possibility to override max capacity of a
        variable length array in serialization routines. This option will disable serialization
        buffer checks and add conditional compilation statements which violates MISRA.

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--language-standard",
        "-std",
        choices=["c11", "c17", "c23", "c++14", "cetl++14-17", "c++17", "c++17-pmr", "c++20", "c++20-pmr"],
        help=textwrap.dedent("""

        For language generators that support different standards of their core language this option
        can be used to optimize the output. For example, C templates may generate slightly different
        code for the the c99 standard then for c11. For available support in Nunavut see the
        documentation for built-in templates
        (https://nunavut.readthedocs.io/en/latest/docs/templates.html#built-in-template-guide).

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--configuration",
        "-c",
        nargs="*",
        type=Path,
        help=textwrap.dedent("""

        There is a set of built-in configuration for Nunavut that provides default values for known
        languages as documented `in the template guide
        <https://nunavut.readthedocs.io/en/latest/docs/templates.html#language-options>`_. This
        argument lets you specify an override configuration json file where any value provided will
        override the built-in defaults. To see the built-in defaults you can use::

            nnvg --list-configuration --list-format json-pretty > built-in.json

        This will generate a json file with all the built-in configuration values as the "configuration.sections"
        value. If you have `jq` installed you can use this to filter the output to just the configuration::

            nnvg --list-configuration --list-format json-pretty | jq '.configuration.sections' > my-config.json

        Then you can edit the my-config.json file to provide overrides for the built-in defaults and use
        this option to provide the file to nnvg::

            nnvg --configuration my-config.json ...


        Also see ``--list-to-file`` which writes this configuration to disk if combined with ``--list-configuration``.

    """).lstrip(),
    )

    ln_opt_group.add_argument(
        "--option",
        "-o",
        nargs="*",
        type=keyvalue,
        help=textwrap.dedent("""
        Passes a key=value pair to the template as a language option overriding default options where they are
        specified. This is useful for passing options to the template that are not available as command-line
        arguments. For example, if you have a template that uses the
        "foo" variable you can pass a value to it using this argument::

            nnvg -o foo=value ...

        where "foo" is then available in the template as `{{ options.foo }}`.

        This option is similar to the `--configuration` option but only applies to the options section of the target
        language and doesn't require a file to be created.

    """).lstrip(),
    )

    return parser


def make_argparse_parser() -> argparse.ArgumentParser:
    """
    Defines the command-line interface using generic argparse.ArgumentParser.
    This should be used for documentation and tab-completion tools.
    """
    return _make_parser(argparse.ArgumentParser)



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
A small collection of common utilities.

.. note::

    Please don't use this as a dumping ground for things that belong in a dedicated package. Python being such a
    full-featured language, there should be very few truly generic utilities in Nunavut.

"""

import collections.abc
import copy
import enum
import logging
import pathlib
import sys
from importlib import resources as importlib_resources
from typing import Any, Callable, Generator, Generic, MutableMapping, Optional, TypeVar, cast

_logger = logging.getLogger(__name__)


TEMPLATE_SUFFIX = ".j2"  #: The suffix expected for nunavut templates.


@enum.unique
class YesNoDefault(enum.Enum):
    """
    Trinary type for decisions that allow a default behavior to be requested that can
    be different based on other contexts. For example:

    .. invisible-code-block: python

        from datetime import datetime
        from nunavut._utilities import YesNoDefault

    .. code-block:: python

        def should_we_order_pizza(answer: YesNoDefault) -> bool:
            if answer == YesNoDefault.YES or (
               answer == YesNoDefault.DEFAULT and
               datetime.today().isoweekday() == 5):
                # if yes or if we are taking the default action which is to
                # order pizza on Friday, and today is Friday, then we order pizza
                return True
            else:
                return False

    .. invisible-code-block: python

        assert should_we_order_pizza(YesNoDefault.YES)
        assert not should_we_order_pizza(YesNoDefault.NO)

    """

    @classmethod
    def test_truth(cls, ynd_value: "YesNoDefault", default_value: bool) -> bool:
        """
        Helper method to test a YesNoDefault value and return a default boolean value.

        .. invisible-code-block: python

            from nunavut._utilities import YesNoDefault

        .. code-block:: python

            '''
                let "is YES" be Y
                let "is DEFAULT" be D where:
                    if Y then not D and if D then not Y
                    and "is NO" is Y = D = 0
                let "is default_value true" be d

                Y | D | d | Y or (D and d)
                1   *   *    1
                0   1   0    0
                0   1   1    1
                0   0   *    0
            '''

            assert YesNoDefault.test_truth(YesNoDefault.YES, False)
            assert not YesNoDefault.test_truth(YesNoDefault.DEFAULT, False)
            assert YesNoDefault.test_truth(YesNoDefault.DEFAULT, True)
            assert not YesNoDefault.test_truth(YesNoDefault.NO, True)

        """
        if ynd_value == cls.DEFAULT:
            return default_value
        else:
            return ynd_value == cls.YES

    NO = 0
    YES = 1
    DEFAULT = 2


@enum.unique
class QuaternaryLogic(enum.Enum):
    """
    A quaternary logic value.

    .. invisible-code-block: python

        from nunavut._utilities import QuaternaryLogic
        from pytest import raises

        assert QuaternaryLogic.from_en_us(str(False)) == QuaternaryLogic.ALWAYS_FALSE
        assert QuaternaryLogic.from_en_us(int(False)) == QuaternaryLogic.ALWAYS_FALSE
        assert QuaternaryLogic.from_en_us(None) == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("default") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us(str(None)) == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us(str(True)) == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us(int(True)) == QuaternaryLogic.TRUE_IF

        assert QuaternaryLogic.from_en_us("always_false") == QuaternaryLogic.ALWAYS_FALSE
        assert QuaternaryLogic.from_en_us("always-false") == QuaternaryLogic.ALWAYS_FALSE
        assert QuaternaryLogic.from_en_us("never") == QuaternaryLogic.ALWAYS_FALSE
        assert QuaternaryLogic.from_en_us("No") == QuaternaryLogic.ALWAYS_FALSE

        assert QuaternaryLogic.from_en_us("true_if") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("true-if") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("as-needed") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("if-needed") == QuaternaryLogic.TRUE_IF
        assert QuaternaryLogic.from_en_us("yes") == QuaternaryLogic.TRUE_IF

        assert QuaternaryLogic.from_en_us("true_unless") == QuaternaryLogic.TRUE_UNLESS
        assert QuaternaryLogic.from_en_us("true-unless") == QuaternaryLogic.TRUE_UNLESS
        assert QuaternaryLogic.from_en_us("only") == QuaternaryLogic.TRUE_UNLESS

        assert QuaternaryLogic.from_en_us("always_true") == QuaternaryLogic.ALWAYS_TRUE
        assert QuaternaryLogic.from_en_us("always-true") == QuaternaryLogic.ALWAYS_TRUE
        assert QuaternaryLogic.from_en_us("always") == QuaternaryLogic.ALWAYS_TRUE

        with raises(ValueError):
            QuaternaryLogic.from_en_us("not_a_value")

    Example usage:

    .. code-block:: python

        def should_we_order_pizza(answer: QuaternaryLogic, is_today_friday: bool) -> bool:
            if answer == QuaternaryLogic.TRUE_IF:
                # order pizza on Friday!
                return is_today_friday
            elif answer == QuaternaryLogic.TRUE_UNLESS:
                # only order pizza if it's not Friday
                return not is_today_friday
            elif answer == QuaternaryLogic.ALWAYS_TRUE:
                # always order pizza
                return True
            elif answer == QuaternaryLogic.ALWAYS_FALSE:
                # never order pizza
                return False
            else:
                raise ValueError("Unknown value")

    """

    @classmethod
    def from_en_us(cls, en_us_word: Any) -> "QuaternaryLogic":
        """
        Convert an English words for "always false, always true, true if, and true unless" to a quaternary logic
        value.

        :param en_us_word: The English word to convert.
        :return: The input as a quaternary logic value.
        :raises ValueError: If the word is not recognized.

        """

        if en_us_word is None:
            return cls.TRUE_IF

        lcw = str(en_us_word).lower()
        if lcw in ("always_false", "always-false", "never", "no", "false", "0"):
            return cls.ALWAYS_FALSE
        if lcw in ("true_if", "true-if", "as-needed", "if-needed", "yes", "", "default", "none", "true", "1"):
            return cls.TRUE_IF
        if lcw in ("true_unless", "true-unless", "only"):
            return cls.TRUE_UNLESS
        if lcw in ("always_true", "always-true", "always"):
            return cls.ALWAYS_TRUE
        raise ValueError(f"Unknown value '{en_us_word}'")

    ALWAYS_FALSE = 0
    """
    Always false.
    """

    ALWAYS_TRUE = 1
    """
    Always true.
    """

    TRUE_IF = 2
    """
    True if a condition is met. (1 AND condition)
    """

    TRUE_UNLESS = 3
    """
    True unless a condition is met. (1 XOR condition)
    """


@enum.unique
class ResourceType(enum.Enum):
    """
    Standard Nunavut classifications for Python package resources.
    """

    NONE = 0
    """ No resources specified."""
    SERIALIZATION_SUPPORT = 0x1
    """Serialization support files."""
    TYPE_SUPPORT = 0x2
    """Type support files."""
    ONLY = 0x80000000
    """Only the specified resources."""
    ANY = 0x3
    """Any resources."""


@enum.unique
class ResourceSearchPolicy(enum.Enum):
    """
    Generic policy type for controlling the behaviour of things that search for resources.
    """

    FIND_ALL = 0
    FIND_FIRST = 1


def iter_package_resources(pkg_name: str, *suffix_filters: str) -> Generator[pathlib.Path, None, None]:
    """
    A generator that yields all the resources in a package that match a given suffix filter.

    Example usage:

    .. invisible-code-block: python

        from nunavut._utilities import iter_package_resources

    .. code-block:: python

        for x in iter_package_resources("nunavut.lang", ".py"):
            print(x)

    .. invisible-code-block: python

        rs = [x for x in iter_package_resources("nunavut.lang", ".py") if x.name == "__init__.py"]
        assert 1 == len(rs)
        assert rs[0].name == '__init__.py'

    """
    for resource in importlib_resources.files(pkg_name).iterdir():
        if resource.is_file() and isinstance(resource, pathlib.Path):
            # Not sure why this works but it's seemed to so far. importlib_resources.as_file(resource)
            # may be more correct but this can create temporary files which would disappear after the iterator
            # had copied their paths. If you are reading this because this method isn't working for some packaging
            # scheme then we may need to use importlib_resources.as_file(resource) to create a runtime cache of
            # temporary objects that live for a given nunavut session. This, of course, wouldn't help across sessions
            # which is a common use case when integrating Nunavut with build systems. So...here be dragons.
            file_resource = resource
            if any(suffix == file_resource.suffix for suffix in suffix_filters):
                yield file_resource


def empty_list_support_files() -> Generator[pathlib.Path, None, None]:
    """
    Helper for implementing the list_support_files method in language support packages. This provides an empty
    iterator with the correct type annotations.
    """
    # works in Python 3.3 and newer. Thanks https://stackoverflow.com/a/13243870
    yield from ()


class DefaultValue:
    """
    Represents a default value in the language configuration. Use this to differentiate between explicit values and
    default values when merging configuration. For example, given the following configuration:

    .. invisible-code-block: python

        from nunavut import DefaultValue

    .. code-block:: python

        collection = {
            'a': DefaultValue(1),
            'b': 2
        }

        overrides = [
            {
                'a': 3,
                'b': DefaultValue(4)
            },
            {
                'a': DefaultValue(5),
                'b': 6
            }
        ]

    Then the merged configuration should be:

    .. code-block:: python

        merged = {
            'a': 3,
            'b': 6
        }

    .. invisible-code-block: python

        # let's try it
        for override in overrides:
            collection = deep_update(collection, override)

        assert collection['a'] == merged['a']
        assert collection['b'] == merged['b']

    Other properties of DefaultValue:

    .. code-block:: python

        assert DefaultValue(1) == 1
        assert DefaultValue(1) != 2
        assert DefaultValue(1) == DefaultValue(1)
        assert DefaultValue(1) != DefaultValue(2)
        assert eval(repr(DefaultValue(1))) == DefaultValue(1)
        assert hash(DefaultValue(1)) == hash(1)
        assert bool(DefaultValue(1))
        assert not bool(DefaultValue(None))
        repred = eval(repr(DefaultValue(8)))
        assert repred.value == 8

    """

    @classmethod
    def assign_to_if_not_default(cls, target: MutableMapping[str, Any], key: str, value: Any) -> Any:
        """
        Assigns a value to a key in a dictionary unless the key already has a value and the value is not a
        `DefaultValue`. The one exception to this is if the value is a `DefaultValue` and the value for the key is
        already a `DefaultValue`. In this case the new `DefaultValue` value will be assigned to the key.

        :param target: The dictionary to assign to.
        :param key: The key to assign to.
        :param value: The value to test and assign.
        :return: The value assigned to the key. This is the value of the `value` parameter if it was assigned or the
                    value of the key in the target dictionary if it was not assigned.
        """
        try:
            if isinstance(value, DefaultValue) and not isinstance(target[key], DefaultValue):
                return target[key]
        except KeyError:
            pass
        target[key] = value
        return value

    def __init__(self, value: Any) -> None:
        self._value = value

    @property
    def value(self) -> Any:
        """
        The default value.
        """
        return self._value

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DefaultValue):
            return bool(self._value == other.value)
        return bool(self._value == other)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"DefaultValue({self.value})"

    def __str__(self) -> str:
        return f"DefaultValue({self.value})"

    def __hash__(self) -> int:
        return hash(self._value)

    def __bool__(self) -> bool:
        return bool(self._value)


def no_default_value(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to convert a function that may return `DefaultValue`s to a function that returns the value of the any
    `DefaultValue`s found. For example:

    .. invisible-code-block: python

        from nunavut._utilities import DefaultValue, no_default_value

    .. code-block:: python

        @no_default_value
        def some_function() -> DefaultValue:
            return DefaultValue(1)

        assert some_function() == 1
        assert not isinstance(some_function(), DefaultValue)
    """

    def wrapper(*args: Any, **kwargs: Any) -> Any:
        result = func(*args, **kwargs)
        if isinstance(result, DefaultValue):
            return result.value
        return result

    return wrapper


DeepUpdateT = TypeVar("DeepUpdateT", bound=MutableMapping)


def deep_update(target: DeepUpdateT, source: DeepUpdateT) -> DeepUpdateT:
    """
    Helper method to do a recursive update of a map that may contain maps as values.

    .. invisible-code-block: python

        from nunavut._utilities import deep_update

    .. code-block:: python

        target_map   =  {
                            "a": { "one": 1, "two": 2 },
                            "b": "not a map"
                        }
        update_from  =  {
                            "a": { "two": { "i": "this value" }, "three": "that value"},
                            "c": "see"
                        }

        target_map   = deep_update(target_map, update_from)
        update_from["a"]["two"]["i"] = "whoops, this was supposed to be a copy"

        assert target_map["a"]["one"] == 1
        assert isinstance(target_map["a"]["two"], collections.abc.Mapping)
        assert target_map["a"]["two"]["i"] == "this value"
        assert target_map["a"]["three"] == "that value"
        assert target_map["b"] == "not a map"
        assert "c" in target_map
        assert target_map["c"] == "see"

    Note that this method is `DefaultValue` aware. If a value in the target map is a `DefaultValue` then it will not
    overwrite the value in the target map. If the value in the source map is a `DefaultValue` then it will not be
    used to update existing values of any type in the target map but will be used to update the target map if the
    target map does not have a value for the given key. In such cases the `DefaultValue` will be inserted into the
    target map.

    .. code-block:: python

        from nunavut import DefaultValue
        target_map   =  {
                            "a": { "one": 1, "two": DefaultValue(2) },
                            "b": "not a default",
                            "c": DefaultValue("one default...")
                        }
        update_from  =  {
                            "a": { "two": { "i": "this value" }, "three": DefaultValue("that value")},
                            "b": DefaultValue("see"),
                            "c": DefaultValue("...deserves another."),
                            "d": DefaultValue("This happened.")
                        }

        target_map   = deep_update(target_map, update_from)

        assert target_map["a"]["one"] == 1
        assert target_map["a"]["two"]["i"] == "this value"
        assert target_map["a"]["three"] == "that value"
        assert target_map["b"] == "not a default"
        assert target_map["c"] == "...deserves another."
        assert target_map["d"] == "This happened."

    """
    if isinstance(target, collections.abc.Mapping):
        for key, value in source.items():
            if isinstance(value, collections.abc.Mapping):
                target[key] = deep_update(target.get(key, {}), cast(DeepUpdateT, value))
            else:
                DefaultValue.assign_to_if_not_default(target, key, value)
    else:
        target = copy.copy(source)
    return target


PropertyT = TypeVar("PropertyT")


class cached_property(Generic[PropertyT]):
    """
    Based on `functools.cached_property` (Python Foundation License 2.0, SPDX: PSF-2.0) implementation in Python 3.11,
    this is both a backport for older Python versions and a version that omits the problematic lock as documented for
    Python 3.12. As such, this version is not thread safe.

    :param func: The function to be wrapped by this decorator.

    .. invisible-code-block: python

        from nunavut._utilities import cached_property
        from pytest import raises

        class Test:

            @classmethod
            @cached_property
            def cls_test(cls) -> int:
                return 1

            def __init__(self) -> None:
                self.calls = 0

            @cached_property
            def test(self) -> int:
                self.calls += 1
                return self.calls

        t = Test()
        assert t.test == 1
        assert t.test == 1
        assert t.test == 1

    """

    _NOT_FOUND = object()

    def __init__(self, func: Callable[..., PropertyT]):
        self._func = func
        self._attr_name: str | None = None
        self.__doc__ = func.__doc__

    def __set_name__(self, owner: Any, name: str) -> None:
        self._attr_name = name

    def __get__(self, instance: Any, owner: Any | None = None) -> PropertyT:
        if self._attr_name is None:  # pragma: no cover
            raise TypeError("Cannot use cached_property instance without calling __set_name__ on it.")
        cache = instance.__dict__
        val = cast(PropertyT, cache.get(self._attr_name, self._NOT_FOUND))
        if val is self._NOT_FOUND:
            val = self._func(instance)
            cache[self._attr_name] = val
        return val



#
# Copyright (C) OpenCyphal Development Team  <opencyphal.org>
# Copyright Amazon.com Inc. or its affiliates.
# SPDX-License-Identifier: MIT
#
"""
Contains supporting C headers to distribute with generated types.
"""

import pathlib
import typing

from nunavut._utilities import ResourceType, empty_list_support_files, iter_package_resources

__version__ = "1.0.0"
"""Version of the c support headers."""


def list_support_files(resource_type: int = ResourceType.ANY.value) -> typing.Generator[pathlib.Path, None, None]:
    """
    Get a list of C support headers embedded in this package.
    :param resource_type: A type of support file to list.

    .. invisible-code-block: python

        from nunavut.lang.c.support import list_support_files
        import pathlib
        support_file_count = 0

    .. code-block:: python

        for path in list_support_files():
            support_file_count += 1
            assert path.parent.stem == 'support'
            assert (path.suffix == '.h' or path.suffix == '.j2')

    .. invisible-code-block: python

        assert support_file_count > 0

    :return: A list of C support header resources.
    """
    # The c support only has serialization support resources
    if 0 == (resource_type & ResourceType.SERIALIZATION_SUPPORT.value):
        return empty_list_support_files()
    else:
        return iter_package_resources(__name__, ".h", ".j2")
